/************************************************************/
/*                        scepstrum.c                       */
/*                                                          */
/*                       SPro Sources                       */
/* -------------------------------------------------------- */
/* Guig                                           Feb. 1999 */
/************************************************************/
/*
 * CVS log:
 *
 * $Author: gravier $
 * $Date: 1999/12/03 13:21:40 $
 * $Revision: 1.5 $
 *
 */
/*
 * Convert speech input file to cepstral coefficients output 
 * param file, using a (real) discrete Fourier transform for
 * a given frequency band.
 *
 * -----
 *
 * Input:  raw, linear 16 bits/sample signal file,
 *         SPHERE signal file
 * Output: SPro FBCEPSTRA data file
 */

#define _scepstrum_c_
#ifndef lint
static char *cvsid = "$Id: scepstrum.c,v 1.5 1999/12/03 13:21:40 gravier Exp $";
#endif

#include <stdio.h>
#include <stdlib.h>

#include <gutil.h>
#include <gerror.h>
#include <errcode.h>
#include <spro.h>

/* ---------------------------- */
/* ----- void usage(void) ----- */
/* ---------------------------- */
void usage(void)
{
  fprintf(stderr,"\n@%s\n",SPRO_V);
  fprintf(stderr,"\nUsage:\n"); 
  fprintf(stderr,"    %s [options] infile outfile\n",prog);
  fprintf(stderr,"\n");
  fprintf(stderr,"Synopsis:\n");
  fprintf(stderr,"    Calculate cepstral coefficients for input speech signal,\n");
  fprintf(stderr,"    using real DFT.\n");
  fprintf(stderr,"\n");
  fprintf(stderr,"Options:\n");
#ifdef SPHERE
  fprintf(stderr,"  --sphere, -S              Sphere signal file format (pcm)\n");
#endif /* SPHERE */
  fprintf(stderr,"  --sample-rate, -f num     signal sample rate (8000 Hz)\n");
  fprintf(stderr,"  --freq-min, -i num        lower frequency bound (0 Hz)\n");
  fprintf(stderr,"  --freq-max, -u num        higher frequency bound (Fs/2)\n");
  fprintf(stderr,"  --pre-emphasis, -k num    pre-emphasis coefficient (0.95)\n");
  fprintf(stderr,"  --length, -l num          frame length in ms (20)\n");
  fprintf(stderr,"  --decay, -d num           frame decay in ms (10)\n");
  fprintf(stderr,"  --window, -w str          weighting window (hamming)\n");
  fprintf(stderr,"  --num-cep, -p num         number of cepstral coefficients (10)\n");
  fprintf(stderr,"  --cms, -z                 cepstral mean removal\n");
  fprintf(stderr,"  --lifter, -r num          liftering value (0)\n");
  fprintf(stderr,"  --energy, -e              add log-energy to ouput file\n");
  fprintf(stderr,"  --scale-energy, -n num    scale energy (off)\n");
  fprintf(stderr,"  --verbose, -v             verbose mode\n");
  fprintf(stderr,"  --version, -V             print version number\n");
  fprintf(stderr,"\n");
}

/* ----------------------------- */
/* ----- Long option array ----- */
/* ----------------------------- */
swtab_t longopts[]={
#ifdef SPHERE
  {"sphere",'S'},
#endif /* SPHERE */
  {"sample-rate",'f'},
  {"pre-emphasis",'k'},
  {"freq-min",'i'},
  {"freq-max",'u'},
  {"length",'l'},
  {"decay",'d'},
  {"window",'w'},
  {"num-cep",'p'},
  {"cms",'z'},
  {"lifter",'r'},
  {"energy",'e'},
  {"scale-energy",'n'},
  {"verbose",'v'},
  {"version",'V'},
  {"help",'h'},
  {NULL,0x00}
};

/* ---------------------------------------------- */
/* ----- Global variables set by ReadArgs() ----- */
/* ---------------------------------------------- */
bool_t trace=false;    /* Set to true with -v option           */
float freq=8000.0;     /* Sample rate of the input signal      */
float f_min=0.0;       /* Lower frequency bound                */
float f_max=0.0;       /* Higher frequency bound               */
float fm_l=20.0;       /* Frame length in ms                   */
float fm_d=10.0;       /* Frame decay in ms                    */
float emphco=0.95;     /* Pre-emphasis coefficient             */
win_t win=HAMMING;     /* Weighting window                     */
bool_t cms=false;      /* Set to true with -z option           */
int numceps=10;        /* Number of cepstral ceofficients      */
int lifter=0;          /* Liftering value                      */
flag_t flag=0;         /* Qualifier flags                      */
float escale=0.0;      /* Energy scaling factor                */
#ifdef SPHERE
bool_t sphere=false;   /* Set to true with -S flag             */
#endif /* SPHERE */

/* --------------------------------- */
/* ----- int main(int,char **) ----- */
/* --------------------------------- */
int main(int argc,char **argv)
{
  char *ifile;           /* Input filename                       */
  char *ofile;           /* Output filename                      */
  FILE *f;               /* I/O file descriptor                  */
  sig_t *sig;            /* Input signal                         */
  data_t *data;          /* Output data structure                */
  err_t status;          /* Return status from FBAnalysis()      */
  int nframe;            /* Number of frames in sig              */
  bool_t ReadArgs(void);
  bool_t CheckArgs(void);
  err_t CepstralAnalysis(sig_t *,data_t *);

  /*
   * Process input arguments
   */
  LongArgInit(argc,argv,longopts);
  if(!ReadArgs()) {
    usage();
    return(ARGS_ERR);
  }
  if(f_max == 0.0)
    f_max=freq/2.0;
  if(!CheckArgs())
    return(ARGS_ERR);
  if(NextArgType() != STRARG) {
    gerror(ARGS_ERR,"%s: no input signal file specified",prog);
    usage();
    return(ARGS_ERR);
  }
  ifile=GetStrArg();
  if(NextArgType() != STRARG) {
    gerror(ARGS_ERR,"%s: no output signal file specified",prog);
    usage();
    return(ARGS_ERR);
  }
  ofile=GetStrArg();
  /*
   * Trace required operations
   */
  if(trace) {
    fprintf(stdout,"\n%s (%s): %s --> %s\n\n",prog,SPRO_V,ifile,ofile);
    fflush(stdout);
  }
  /*
   * Load input signal
   */
#ifdef SPHERE
  if(sphere) {
    if((status=ReadSphereSig(&sig,&freq,ifile)) != OK) {
      gerror(status,"%s: cannot load signal from file %s",prog,ifile);
      return(status);
    }
  }
  else {
#endif /* SPHERE */
    if((f=fopen(ifile,"rb")) == NULL) {
      gerror(FOPEN_ERR,"%s: cannot open input file %s",prog,ifile);
      return(FOPEN_ERR);
    }
    if((status=ReadSig(&sig,fsize(f)/2,f)) != OK) {
      gerror(status,"%s: cannot load signal from file %s",prog,ifile);
      fclose(f);
      return(status);
    }
    fclose(f);
#ifdef SPHERE
  }
#endif /* SPHERE */
  /*
   * Run pre-emphasis
   */
  (void)SigEmph(sig,emphco);
  /*
   * Get memory for the output structure
   */
  nframe=1+(sig->n-(int)(fm_l*freq/1000.0))/(int)(fm_d*freq/1000.0);
  if((status=CreateSPData(&data,FBCEPSTRA,numceps,nframe,flag)) != OK) {
    gerror(status,"%s: cannot create output data structure",prog);
    FreeSig(sig);
    return(status);
  }
  /*
   * Run cepstral analysis
   */
  if((status=CepstralAnalysis(sig,data)) != OK) {
    gerror(status,"%s: fatal error",prog);
    FreeSig(sig);
    FreeSPData(data);
    return(status);
  }
  if(cms)
    DataZeroMean(data);
  if(escale != 0.0)
    ScaleEnergy(data,escale);
  /*
   * Save data to output file
   */
  if((f=fopen(ofile,"wb")) == NULL) {
    gerror(FOPEN_ERR,"%s: cannot open output file %s",prog,ofile);
    FreeSig(sig);
    FreeSPData(data);
    return(FOPEN_ERR);
  }
  if((status=WriteSPData(data,f)) != OK) {
    gerror(status,"%s: cannot dump data to output file %s",prog,ofile);
    FreeSig(sig); FreeSPData(data);
    return(status);
  }
  fclose(f);

  FreeSig(sig);
  FreeSPData(data);

  DONE;
}

/* ----------------------------------------------------- */
/* ----- err_t CepstralAnalysis(sig_t *,data_t *) ------ */
/* ----------------------------------------------------- */
/*
 * Here is the main standard loop for cepstral analysis.
 * Roughly, it does:
 *  - get a frame of signal
 *  - pre-process it
 *  - run FFT on it
 *  - extract the [f_min,f_max] part of it
 *  - make it symetric
 *  - apply real DFT algorithm
 *  - apply liftering (if any)
 *  - write CC to data structure
 *  - write energy (if any) to data structure
 */
err_t CepstralAnalysis(sig_t *sig,data_t *data)
{
  int i,j;               /* Some loopers                         */
  int M,N;               /* f_min (resp f_max) bin in FFT vector */
  int nelem;             /* real output vector length            */
  sig_t *frame;          /* A frame of signal                    */
  vector_t *v1;          /* Full-band FFT log-module (non-sym)   */
  vector_t *v2;          /* Band-passed FFT log-module           */
  vector_t *v3;          /* Cesptral coefficients                */
  float c0;              /* C0 (or energy) returned by dft()     */
  err_t status;          /* Return status from inferior func.    */
  
  void fft_extract(vector_t *,int,int,vector_t *);
  float dft(vector_t *,int,vector_t *);

  /*
   * Compute M, N and L constants
   */
  M=round(FBANK_FFT_SIZE*f_min/freq);
  if((N=round(FBANK_FFT_SIZE*f_max/freq)) >= FBANK_FFT_SIZE/2)
    N=FBANK_FFT_SIZE/2-1;

  /*  fprintf(stderr,"M= %d, N= %d\n",M,N); */
  /*
   * Get required memory
   */
  if((status=CreateSig(&frame,(unsigned long)(fm_l*freq/1000.0))) != OK)
    return(status);
  if((status=CreateVector(&v1,FBANK_FFT_SIZE/2)) != OK) {
    FreeSig(frame);
    return(status);
  }
  if((status=CreateVector(&v2,2*(N-M+1))) != OK) {
    FreeSig(frame); FreeVector(v1);
    return(status);
  }
  if((status=CreateVector(&v3,numceps)) != OK) {
    FreeSig(frame); FreeVector(v1); FreeVector(v2);
    return(status);
  }
  nelem=DataVecSize(data);  
  /*
   * Process every frame
   */
  for(i=0;i < data->n;i++) {
    /*
     * Get and pre-process frame
     */
    if((status=SigGetFrame(sig,i+1,frame->n,(int)(fm_d*freq/1000.0),frame)) != OK) {
      gerror(status,"CepstralAnalysis(): error reading frame %d",prog,i);
      FreeSig(frame); FreeVector(v1); FreeVector(v2); FreeVector(v3);
      return(status);
    }
    SigWeight(frame,win);
    /*
     * Apply FFT
     */
    if((status=FFTransform(frame,FBANK_FFT_SIZE,v1,NULL)) != OK) {
      gerror(SPRO_ERR,"CepstralAnalysis(): cannot compute FFT for frame %d\n",i);
      FreeSig(frame); FreeVector(v1); FreeVector(v2); FreeVector(v3);
      return(status);
    }
    /*
     * Extract the meaningfull part and run DFT algorithm
     */
    fft_extract(v1,M,N,v2);
    c0=dft(v2,numceps,v3);
    /*
     * Apply all necessary stuff
     */
    if(lifter != 0)
      CepLifter(v3,lifter);
    /*
     * Copy parameter vector to data structure
     */
    for(j=0;j < data->dim;j++)
      data->s[i*nelem+j]=v3->data[j];
    if(data->flag & WITHE)
      data->s[i*nelem+j]=c0 < ENERGY_FLOOR ? (param_t)(2.0*log(ENERGY_FLOOR)) : (param_t)(2.0*log(c0));
  }
  /*
   * Give memory back and say OK
   */
  FreeSig(frame);
  FreeVector(v1);
  FreeVector(v2);
  FreeVector(v3);

  DONE;
}

/* ----------------------------------------------------------- */
/* ----- void fft_extract(vector_t *,int,int,vector_t *) ----- */
/* ----------------------------------------------------------- */
void fft_extract(vector_t *v1,int M,int N,vector_t *v2)
{
  int i;

  v2->data[0]=log(v1->data[M]);

  for(i=1;i < N-M+1;i++) {
    v2->data[i]=log(v1->data[M+i]);
    v2->data[v2->n-i]=log(v1->data[M+i]);
  }
  v2->data[i]=0.0;
}

/* ----------------------------------------------- */
/* ----- void dft(vector_t *,int,vector_t *) ----- */
/* ----------------------------------------------- */
/*
 *        v1                    v2
 * x(0) ----- x(N-1) --> X(0) ------ X(p)
 *
 * X(0) = \sum_(0}^{N-1} x(n)
 * X(k) = \sum_{0}^{N-1} (x(n) cos(2 i \pi k n/N))
 *
 * Since we are dft'ing real symetric spectra, it's no use computing
 * the imaginary part of the transform which is null.
 */
float dft(vector_t *v1,int p,vector_t *v2)
{
  int k,n;
  float c0=0.0;

  for(n=0;n < v1->n;n++)
    c0 += v1->data[n];

  for(k=1;k <= p;k++) {
    v2->data[k-1]=0.0;
    for(n=0;n < v1->n;n++)
      v2->data[k-1] += (v1->data[n] * cos(2.0*M_PI*(float)k*(float)n/(float)(v1->n)));
  }
  
  return(c0);
}

/* ---------------------------------- */
/* ----- bool_t CheckArgs(void) ----- */
/* ---------------------------------- */
bool_t CheckArgs(void)
{
  if(f_min < 0 || f_min >= f_max || f_max > freq/2) {
    gerror(SPRO_ERR,"%s: invalid frequency range [%.2f,%.2f]\n",prog,f_min,f_max);
    return(false);
  }
  if(numceps < 1 || numceps > MAX_CEPSTRUM) {
    gerror(SPRO_ERR,"%s: invalid number of CC %d (min= 1, max= %d)",
	   prog,numceps,MAX_CEPSTRUM);
    return(false);
  }
  return(true);
}

/* --------------------------------- */
/* ----- bool_t ReadArgs(void) ----- */
/* --------------------------------- */
/*
 *  --sphere, -S              Sphere signal file format (pcm)
 *  --sample-rate, -f num     signal sample rate (8000 Hz)
 *  --pre-emphasis, -k num    pre-emphasis coefficient (0.95)
 *  --freq-min, -i num        lower frequency bound (0 Hz)
 *  --freq-max, -m num        higher frequency bound (Fs/2)
 *  --length, -l num          frame length in ms (20)
 *  --decay, -d num           frame decay in ms (10)
 *  --window, -w str          weighting window (hamming)
 *  --num-cep, -p num         number of cepstral coefficients (10)
 *  --cms, -z                 cepstral mean removal
 *  --lifter, -r num          liftering value (0)
 *  --energy, -e              add log-energy to ouput file
 *  --scale-energy, -n num    scale log-energy
 *  --verbose, -v             verbose mode
 *  --version, -V             print version number
 */
bool_t ReadArgs(void)
{
  char opt;
  char *lwin;
  
  while((opt=GetSwitch()) != NOSWT)
    switch(opt) {
#ifdef SPHERE
    case 'S':
      sphere=true;
      break;
#endif /* SPHERE */
     case 'f':
      if(NextArgType() != NUMARG) {switcherr('f');}
      freq=GetNumArg();
      break;
    case 'k':
      if(NextArgType() != NUMARG) {switcherr('k');}
      emphco=GetNumArg();
      break;
    case 'i':
      if(NextArgType() != NUMARG) {switcherr('i');}
      f_min=GetNumArg();
      break;
    case 'u':
      if(NextArgType() != NUMARG) {switcherr('i');}
      f_max=GetNumArg();
      break;
    case 'l':
      if(NextArgType() != NUMARG) {switcherr('l');}
      fm_l=GetNumArg();
      break;
    case 'd':
      if(NextArgType() != NUMARG) {switcherr('d');}
      fm_d=GetNumArg();
      break;
    case 'w':
      if(NextArgType() != STRARG) {switcherr('w');}
      lwin=GetStrArg();
      if(strcasecmp(lwin,"none") == 0)
	win=RECTANGLE;
      else if(strcasecmp(lwin,"hamming") == 0)
	win=HAMMING;
      else if(strcasecmp(lwin,"hanning") == 0)
	win=HANNING;
      else if(strcasecmp(lwin,"blackman") == 0)
	win=BLACKMAN;
      else {switcherr('w');}
      break;
    case 'p':
      if(NextArgType() != NUMARG) {switcherr('p');}
      numceps=(int)GetNumArg();
      break;
    case 'z':
      cms=true;
      break;
    case 'r':
      if(NextArgType() != NUMARG) {switcherr('r');}
      lifter=(int)GetNumArg();
      break;
    case 'e':
      flag |= WITHE;
      break;
    case 'n':
      if(NextArgType() != NUMARG) {switcherr('n');}
      escale=GetNumArg();
      break;
    case 'v':
      trace=true;
      break;
    case 'V':
      fprintf(stdout,"\n@%s -- %s\n\n",SPRO_V,cvsid);
      exit(0);
      break;
    case 'h':
      return(false);
      break;
    case '-':
      return(false); break;      
    default:
      gerror(ARGS_ERR,"Unknown switch -- %c",opt);
      return(false);
    }

  return(true);
}

#undef _scepstrum_c_
