/************************************************************/
/*                          scopy.c                         */
/*                                                          */
/*                       SPro Sources                       */
/* -------------------------------------------------------- */
/* Guig                                           Apr. 1997 */
/************************************************************/
/*
 * CVS log:
 *
 * $Author: gravier $
 * $Date: 1999/12/02 16:54:31 $
 * $Revision: 1.4 $
 *
 */
/*
 * Copy a data file and makes possible convertions
 *
 * -----
 *
 * Input:  SPro data file
 * Output: SPro data file, or
 *         HTK v1 data file, or
 *         HTK v2 OTHER data file, or
 *         ASCII data file (gnuplot or matlab use)
 */

#define _scopy_c_
#ifndef lint
static char *cvsid = "$Id: scopy.c,v 1.4 1999/12/02 16:54:31 gravier Exp $";
#endif

#include <stdio.h>
#include <stdlib.h>

#include <gutil.h>
#include <gerror.h>
#include <errcode.h>
#include <spro.h>

/* ---------------------------- */
/* ----- void usage(void) ----- */
/* ---------------------------- */
void usage(void)
{
  fprintf(stderr,"\n@%s\n",SPRO_V);
  fprintf(stderr,"\nUsage:\n"); 
  fprintf(stderr,"    %s [options] ifile ofile\n",prog);
  fprintf(stderr,"\n");
  fprintf(stderr,"Synopsis:\n");
  fprintf(stderr,"    Copy ifile or part of ifile to ofile after possible convertion\n");
  fprintf(stderr,"\n");
  fprintf(stderr,"Options:\n");
  fprintf(stderr,"  --start, -s num           starting at sample number (1)\n");
  fprintf(stderr,"  --end, -e num             ending at sample number (last)\n");
  fprintf(stderr,"  --convert, -C str         convertion string - ADNZ (none)\n");
  fprintf(stderr,"  --output-format, -o str   set output format - ascii | htk1 | htk2 (spro)\n");
  fprintf(stderr,"  --verbose, -v             verbose mode\n");
  fprintf(stderr,"  --version, -V             print version number\n");
  fprintf(stderr,"\n");
}

/* ----------------------------- */
/* ----- Long option array ----- */
/* ----------------------------- */
swtab_t longopts[]={
  {"start",'s'},
  {"end",'e'},
  {"convert",'C'},
  {"output-format",'o'},
  {"verbose",'v'},
  {"version",'V'},
  {"help",'h'},
  {NULL,0x00}
};

/* ---------------------------------------------- */
/* ----- Global variables set by ReadArgs() ----- */
/* ---------------------------------------------- */
unsigned long st=1;       /* Start at sample st (default: first)  */
unsigned long en=0;       /* Start at sample en (default: last)   */
bool_t trace=false;       /* verbose option                       */
char *ofmt=NULL;
flag_t cvtflag=0;         /* Qualifier convertion string          */

/* --------------------------------- */
/* ----- int main(int,char **) ----- */
/* --------------------------------- */
int main(int argc,char **argv)
{
  char *ifile;            /* Input filename                       */
  char *ofile;            /* Output filename                      */
  FILE *f;                /* I/O file descriptor                  */
  data_t *data;           /* Input data structure                 */
  int nelem;
  err_t status;

  bool_t ReadArgs(void);  /* Parse options on command line        */
  bool_t CheckArgs(void); /* Check option validity                */
  void ShowFlag(flag_t);  /* Print qualifiers                     */
  extern void GnuData(data_t *,FILE *);

  /***** Process input arguments *****/
  LongArgInit(argc,argv,longopts);
  if(!ReadArgs()) {
    usage();
    return(ARGS_ERR);
  }
  if(!CheckArgs())
    return(ARGS_ERR);
  if(NextArgType() != STRARG) {
    gerror(ARGS_ERR,"%s: no input file specified",prog);
    usage();
    return(ARGS_ERR);
  }
  ifile=GetStrArg();
  if(NextArgType() != STRARG) {
    gerror(ARGS_ERR,"%s: no output file specified",prog);
    usage();
    return(ARGS_ERR);
  }
  ofile=GetStrArg();

  /***** Show what was asked to do *****/
  if(trace) {
    fprintf(stdout,"\n%s (%s): %s --> %s\n\n",prog,SPRO_V,ifile,ofile);
    fflush(stdout);
  }

  /***** Load data from input file *****/
  if((f=fopen(ifile,"rb")) == NULL) {
    gerror(FOPEN_ERR,"%s: cannot open file %s",prog,ifile);
    return(FOPEN_ERR);
  }
  if((status=LoadSPData(&data,cvtflag,f)) != OK) {
    gerror(status,"%s: cannot load data from file %s",prog,ifile);
    fclose(f);
    return(status);
  }
  fclose(f);
 
  /***** Extract the samples between st and en *****/
  /*
   * To extract the samples between st and en, simply
   * change data->n and data->s so that it points to
   * the st'th vector.
   */

  if(en == 0)
    en=data->n;
  if(en > data->n || st > data->n) {
    gerror(RANGE_ERR,"%s error: invalid start and end values (st= %ul, en= %ul, nsamples= %ul)",
	   prog,st,en,data->n);
    FreeSPData(data);
    return(RANGE_ERR);
  }
  data->n=(en-st)+1;
  if(st != 1) {
    nelem=DataVecSize(data);
    data->s += nelem*(st-1);
  }

  /***** Save to output file *****/
  if((f=fopen(ofile,"wb")) == NULL) {
    gerror(FOPEN_ERR,"%s: cannot open output file %s",prog,ofile);
    FreeSPData(data);
    return(FOPEN_ERR);
  }
  if(ofmt == (char *)NULL) {
    if((status=WriteSPData(data,f)) != OK) {
      gerror(status,"%s: cannot dump data to output file %s",prog,ofile);
      FreeSPData(data); fclose(f);
      return(status);
    }
  }
  else if(strcasecmp(ofmt,"htk1") == 0) {
    gerror(SPRO_WARN,"%s: assuming frame rate of 10 ms in HTK data file %s",
 	   prog,ofile);
    if((status=HTKWrite(data,10.0,1,f)) != OK) {
      gerror(status,"%s: cannot dump data to output file %s",prog,ofile);
      FreeSPData(data); fclose(f);
      return(status);
    }
  }
  else if(strcasecmp(ofmt,"htk2") == 0) {
    gerror(SPRO_WARN,"%s: assuming frame rate of 10 ms in HTK data file %s",
 	   prog,ofile);
    if((status=HTKWrite(data,10.0,2,f)) != OK) {
      gerror(status,"%s: cannot dump data to output file %s",prog,ofile);
      FreeSPData(data); fclose(f);
      return(status);
    }
  }
  else if(strcasecmp(ofmt,"ascii") == 0)
    GnuData(data,f);
  else {
    gerror(SPRO_ERR,"%s: unknown output format -- %s",prog,ofmt);
    FreeSPData(data); fclose(f);
    return(SPRO_ERR);
  }

  /***** House keeping *****/
  fclose(f);
  FreeSPData(data);

  DONE;
}

/* --------------------------------- */
/* ----- void ShowFlag(flag_t) ----- */
/* --------------------------------- */
/*
 * Display qualifier flags
 */
void ShowFlag(flag_t flag)
{
  if(flag & WITHE) fprintf(stdout," E");
  if(flag & WITHD) fprintf(stdout," D");
  if(flag & WITHA) fprintf(stdout," A");
  if(flag & WITHN) fprintf(stdout," N");
  if(flag & WITHZ) fprintf(stdout," Z");
}

/* ---------------------------------- */
/* ----- bool_t CheckArgs(void) ----- */
/* ---------------------------------- */
bool_t CheckArgs(void)
{
  if(en && st > en) {
    gerror(ARGS_ERR,"%s: start sample greater than end sample (st=%d, en=%d)",
	   prog,st,en);
    return(false);
  }
  if(st < 1 || (en && en < 1)) {
    gerror(ARGS_ERR,"%s: invalid start and/or end sample (st=%d, en=%d)",
	   prog,st,en);
    return(false);
  }
  return(true);
}

/* --------------------------------- */
/* ----- bool_t ReadArgs(void) ----- */
/* --------------------------------- */
/*
 *  --start, -s num           starting at sample number (1)\n");
 *  --end, -e num             ending at sample number (last)\n");
 *  --convert, -C str         convertion string - ADNZ (none)\n");
 *  --output-format, -o str   set output format - ascii | htk1 | htk2 (spro)\n");
 *  --verbose, -v             verbose mode\n");
 *  --version, -V             print version number\n");
 *  --help
 */
bool_t ReadArgs(void)
{
  char opt;

  while((opt=GetSwitch()) != NOSWT)
    switch(opt) {
    case 's':
      if(NextArgType() != NUMARG) {switcherr('s');}
      st=(unsigned long)GetNumArg();
      break;
    case 'e':
      if(NextArgType() != NUMARG) {switcherr('e');}
      en=(unsigned long)GetNumArg();
      break;
    case 'C':
      if(NextArgType() != STRARG) {switcherr('C');}
      cvtflag=ConvStr2Flag(GetStrArg());
      break;
    case 'o':
      if(NextArgType() != STRARG) {switcherr('o');}
      ofmt=GetStrArg();
      break;
    case 'v':
      trace=true;
      break;
    case 'V':
      fprintf(stdout,"\n@%s -- %s\n\n",SPRO_V,cvsid);
      exit(0);
      break;
    case 'h':
      return(false);
      break;
    case '-':
      return(false); break;      
    default:
      gerror(ARGS_ERR,"Unknown option switch -- %c",opt);
      return(false);
    }
  return(true);
}
      
#undef _scopy_c_
