/************************************************************/
/*                         sfbcep.c                         */
/*                                                          */
/*                       SPro Sources                       */
/* -------------------------------------------------------- */
/* Guig                                           Apr. 1997 */
/************************************************************/
/*
 * CVS log:
 *
 * $Author: gravier $
 * $Date: 1999/12/03 13:23:42 $
 * $Revision: 1.5 $
 *
 */
/*
 * Convert speech input file to cepstral coefficients output 
 * param file, using filter bank analysis techniques.
 *
 * It allows spectral scale deformation (MEL or ALPHA), cepstral
 * mean removal and liftering. Log energy can also be added to
 * the output file.
 *
 * NOTE: next version/release should deal with script files, so that 
 *       filter bank cut-off index arrays are not re-computed each
 *       time this program is ran. Anyway, the new fast ThetaInv()
 *       procedure should give better performances.
 *
 * -----
 *
 * Input:  raw, linear 16 bits/sample signal file
 * Output: SPro FBCEPSTRA data file
 */

#define _sfbcep_c_
#ifndef lint
static char *cvsid = "$Id: sfbcep.c,v 1.5 1999/12/03 13:23:42 gravier Exp $";
#endif

#include <stdio.h>
#include <stdlib.h>

#include <gutil.h>
#include <gerror.h>
#include <errcode.h>
#include <spro.h>

/* ---------------------------- */
/* ----- void usage(void) ----- */
/* ---------------------------- */
void usage(void)
{
  fprintf(stderr,"\n@%s\n",SPRO_V);
  fprintf(stderr,"\nUsage:\n"); 
  fprintf(stderr,"    %s [options] infile outfile\n",prog);
  fprintf(stderr,"\n");
  fprintf(stderr,"Synopsis:\n");
  fprintf(stderr,"    Calculate cepstral coefficients for input speech signal,\n");
  fprintf(stderr,"    using filter bank analysis\n");
  fprintf(stderr,"\n");
  fprintf(stderr,"Options:\n");
#ifdef SPHERE
  fprintf(stderr,"  --sphere, -S              Sphere signal file format (pcm)\n");
#endif /* SPHERE */
  fprintf(stderr,"  --sample-rate, -f num     signal sample rate (8000 Hz)\n");
  fprintf(stderr,"  --freq-min, -i num        lower frequency bound (0 Hz)\n");
  fprintf(stderr,"  --freq-max, -u num        higher frequency bound (Fs/2)\n");
  fprintf(stderr,"  --pre-emphasis, -k num    pre-emphasis coefficient (0.95)\n");
  fprintf(stderr,"  --length, -l num          frame length in ms (20)\n");
  fprintf(stderr,"  --decay, -d num           frame decay in ms (10)\n");
  fprintf(stderr,"  --window, -w str          weighting window (hamming)\n");
  fprintf(stderr,"  --alpha, -a num           frequency warping parameter (0.0)\n");
  fprintf(stderr,"  --mel, -m                 use MEL frequency scale (off)\n");
  fprintf(stderr,"  --num-cep, -p num         number of cepstral coefficients (16)\n");
  fprintf(stderr,"  --cms, -z                 cepstral mean removal\n");
  fprintf(stderr,"  --lifter, -r num          liftering value (0)\n");
  fprintf(stderr,"  --energy, -e              add log-energy to ouput file\n");
  fprintf(stderr,"  --scale-energy, -s num    scale energy (off)\n");
  fprintf(stderr,"  --verbose, -v             verbose mode\n");
  fprintf(stderr,"  --version, -V             print version number\n");
  fprintf(stderr,"\n");
}

/* ----------------------------- */
/* ----- Long option array ----- */
/* ----------------------------- */
swtab_t longopts[]={
#ifdef SPHERE
  {"sphere",'S'},
#endif /* SPHERE */
  {"sample-rate",'f'},
  {"freq-min",'i'},
  {"freq-max",'u'},
  {"pre-emphasis",'k'},
  {"length",'l'},
  {"decay",'d'},
  {"window",'w'},
  {"num-filter",'n'},
  {"alpha",'a'},
  {"mel",'m'},
  {"num-cep",'p'},
  {"cms",'z'},
  {"lifter",'r'},
  {"energy",'e'},
  {"scale-energy",'s'},
  {"verbose",'v'},
  {"version",'V'},
  {"help",'h'},
  {NULL,0x00}
};

/* ---------------------------------------------- */
/* ----- Global variables set by ReadArgs() ----- */
/* ---------------------------------------------- */
bool_t trace=false;    /* Set to true with -v option           */
float freq=8000.0;     /* Sample rate of the input signal      */
int nfilt=24;          /* Number of filters in the bank        */
float fm_l=20.0;       /* Frame length in ms                   */
float fm_d=10.0;       /* Frame decay in ms                    */
float alpha=0.0;       /* Frequency deformation parameter      */
float emphco=0.95;     /* Pre-emphasis coefficient             */
win_t win=HAMMING;     /* Weighting window                     */
bool_t useMel=false;   /* Wether to use Mel scale or not       */
bool_t alphaSet=false; /* Set to true when -a used             */
bool_t cms=false;      /* Set to true with -z option           */
int numceps=16;        /* Number of cepstral ceofficients      */
int lifter=0;          /* Liftering value                      */
flag_t flag=0;         /* Qualifier flags                      */
float escale=0.0;      /* Energy scaling factor                */
#ifdef SPHERE
bool_t sphere=false;   /* Set to true with -S flag             */
#endif /* SPHERE */
float f_min=0.0;       /* Lower frequency bound                */
float f_max=0.0;       /* Higher frequency bound               */

/* --------------------------------- */
/* ----- int main(int,char **) ----- */
/* --------------------------------- */
int main(int argc,char **argv)
{
  char *ifile;           /* Input filename                       */
  char *ofile;           /* Output filename                      */
  FILE *f;               /* I/O file descriptor                  */
  sig_t *sig;            /* Input signal                         */
  data_t *data;          /* Output data structure                */
  err_t status;          /* Return status from FBAnalysis()      */
  int nframe;            /* Number of frames in sig              */
  bool_t ReadArgs(void);
  bool_t CheckArgs(void);
  err_t CepstralAnalysis(sig_t *,data_t *);

  /***** Process input arguments *****/
  LongArgInit(argc,argv,longopts);
  if(!ReadArgs()) {
    usage();
    return(ARGS_ERR);
  }
  if(!CheckArgs())
    return(ARGS_ERR);
  if(NextArgType() != STRARG) {
    gerror(ARGS_ERR,"%s: no input signal file specified",prog);
    usage();
    return(ARGS_ERR);
  }
  ifile=GetStrArg();
  if(NextArgType() != STRARG) {
    gerror(ARGS_ERR,"%s: no output signal file specified",prog);
    usage();
    return(ARGS_ERR);
  }
  ofile=GetStrArg();

  /***** Show what was asked to do *****/
  if(trace) {
    fprintf(stdout,"\n%s (%s): %s --> %s\n\n",prog,SPRO_V,ifile,ofile);
    fflush(stdout);
  }

  /***** Load input signal *****/
#ifdef SPHERE
  if(sphere) {
    if((status=ReadSphereSig(&sig,&freq,ifile)) != OK) {
      gerror(status,"%s: cannot load signal from file %s",prog,ifile);
      return(status);
    }
  }
  else {
#endif /* SPHERE */
    if((f=fopen(ifile,"rb")) == NULL) {
      gerror(FOPEN_ERR,"%s: cannot open input file %s",prog,ifile);
      return(FOPEN_ERR);
    }
    if((status=ReadSig(&sig,fsize(f)/2,f)) != OK) {
      gerror(status,"%s: cannot load signal from file %s",prog,ifile);
      fclose(f);
      return(status);
    }
    fclose(f);
#ifdef SPHERE
  }
#endif /* SPHERE */

  /***** Run pre-emphasis *****/
  (void)SigEmph(sig,emphco);

  /***** Get memory for the output structure *****/
  nframe=1+(sig->n-(int)(fm_l*freq/1000.0))/(int)(fm_d*freq/1000.0);
  if((status=CreateSPData(&data,FBCEPSTRA,numceps,nframe,flag)) != OK) {
    gerror(status,"%s: cannot create output data structure",prog);
    FreeSig(sig);
    return(status);
  }

  /***** Run filter bank analysis *****/
  if((status=CepstralAnalysis(sig,data)) != OK) {
    gerror(status,"%s: fatal error",prog);
    FreeSig(sig);
    FreeSPData(data);
    return(status);
  }
  if(cms)
    DataZeroMean(data);
  if(escale != 0.0)
    ScaleEnergy(data,escale);
   
  /***** Save data to output file *****/
  if((f=fopen(ofile,"wb")) == NULL) {
    gerror(FOPEN_ERR,"%s: cannot open output file %s",prog,ofile);
    FreeSig(sig);
    FreeSPData(data);
    return(FOPEN_ERR);
  }
  if((status=WriteSPData(data,f)) != OK) {
    gerror(status,"%s: cannot dump data to output file %s",prog,ofile);
    FreeSig(sig); FreeSPData(data);
    return(status);
  }
  fclose(f);

  FreeSig(sig);
  FreeSPData(data);

  DONE;
}

/* ---------------------------------------------------- */
/* ----- err_t CepstralAnalysis(sig_t *,data_t *) ----- */
/* ---------------------------------------------------- */
/*
 * Here is the main standard loop for cepstral analysis.
 * Roughly, it does:
 *  - get a frame of signal
 *  - pre-process it
 *  - apply filter bank analysis
 *  - apply DCT
 *  - apply liftering (if any)
 *  - write CC to data structure
 *  - write log energy (if any) to data structure
 */
err_t CepstralAnalysis(sig_t *sig,data_t *data)
{
  int i,j;
  int nelem;
  int nframe=0;       /* True number of frames since some are suppressed */
  float energy=0.0;
  sig_t *frame;
  vector_t *v1,*v2;
  err_t status;

  /***** Get required memory *****/
  if((status=CreateSig(&frame,(unsigned long)(fm_l*freq/1000.0))) != OK)
    return(status);
  if((status=CreateVector(&v1,nfilt)) != OK) {
    FreeSig(frame);
    return(status);
  }
  if((status=CreateVector(&v2,numceps)) != OK) {
    FreeSig(frame);
    FreeVector(v1);
    return(status);
  }
  nelem=DataVecSize(data);

  /***** Process every frame  *****/
  for(i=1;i<=data->n;i++) {
    /*
     * Get and pre-process frame
     */
    if((status=SigGetFrame(sig,i,frame->n,(int)(fm_d*freq/1000.0),frame)) != OK) {
      gerror(status,"CepstralAnalysis(): error reading frame %d",prog,i);
      FreeSig(frame); FreeVector(v1); FreeVector(v2);
      return(status);
    }
    SigWeight(frame,win);
    /*
     * Apply the transformation
     */
    if(useMel) {
      if((status=LogMelFilterBank(frame,freq,nfilt,f_min,f_max,v1)) != OK) {
	gerror(status,"CepstralAnalysis(): cannot apply MEL LFB to frame %d",prog,i);
	FreeSig(frame); FreeVector(v1); FreeVector(v2);
	return(status);
      }
    }
    else if((status=LogFilterBank(frame,alpha,nfilt,f_min/freq,f_max/freq,v1)) != OK) {
      gerror(status,"%s: cannot apply LFB to frame %d",prog,i);
      FreeSig(frame); FreeVector(v1); FreeVector(v2);
      return(status);
    }
    if((status=DCTransform(v1,numceps,v2)) != OK) {
      gerror(status,"%s: cannot apply DCT to frame %d",prog,i);
      FreeSig(frame); FreeVector(v1); FreeVector(v2);
      return(status);
    }  
    if(lifter != 0)
      CepLifter(v2,lifter);
    if(data->flag & WITHE)
      energy=SigNorm(frame,false);
    /*
     * Copy parameter vector to data
     */
    nframe++;
    for(j=0;j<data->dim;j++)
      data->s[(nframe-1)*nelem+j]=v2->data[j];
    if(data->flag & WITHE)
      data->s[nframe*nelem-1]=energy < ENERGY_FLOOR ?
	(param_t)(2.0*log(ENERGY_FLOOR)) : (param_t)(2.0*log(energy)); /* since energy=sqrt(sum of s[i]*s[i]) */
  }
  data->n=nframe;

  /**** Give memory back and say OK *****/
  FreeSig(frame);
  FreeVector(v1);
  FreeVector(v2);

  DONE;
}

/* ---------------------------------- */
/* ----- bool_t CheckArgs(void) ----- */
/* ---------------------------------- */
bool_t CheckArgs(void)
{
  if(nfilt < MIN_BANK_FILTER || nfilt > MAX_BANK_FILTER) {
    gerror(ARGS_ERR,"%s: invalid number of filters %d (min= %d, max= %d)",
	    prog,nfilt,MIN_BANK_FILTER,MAX_BANK_FILTER);
    return(false);
  }
  if(numceps < 1 || numceps > MAX_CEPSTRUM) {
    fprintf(stderr,"%s: invalid number of CC %d (min= 1, max= %d)",
	    prog,numceps,MAX_CEPSTRUM);
    return(false);
  }
  if(numceps > nfilt) {
    fprintf(stderr,"%s: number of CC > number of filters\n",
	    prog);
    return(false);
  }
  if(alphaSet && useMel)
    gerror(SPRO_WARN,"%s: -a used along with -m (ignoring -a switch)\n",
	   prog);
  return(true);
}

/* --------------------------------- */
/* ----- bool_t ReadArgs(void) ----- */
/* --------------------------------- */
/*
 *  --sphere, -S              Sphere signal file format (pcm)
 *  --sample-rate, -f num     signal sample rate (8000 Hz)
 *  --freq-min, -i num        lower frequency bound (0 Hz)
 *  --freq-max, -u num        higher frequency bound (Fs/2)
 *  --pre-emphasis, -k num    pre-emphasis coefficient (0.95)
 *  --length, -l num          frame length in ms (20)
 *  --decay, -d num           frame decay in ms (10)
 *  --window, -w str          weighting window (hamming)
 *  --num-filter, -n num      number of filters in the filter-bank (24)
 *  --alpha, -a num           frequency warping parameter (0.0)
 *  --mel, -m                 use MEL frequency scale (off)
 *  --num-cep, -p num         number of cepstral coefficients (10)
 *  --cms, -z                 cepstral mean removal
 *  --lifter, -r num          liftering value (0)
 *  --energy, -e              add log-energy to ouput file
 *  --scale-energy, -s num    scale energy (off)
 *  --verbose, -v             verbose mode
 *  --version, -V             print version number
 *  --help
 */
bool_t ReadArgs(void)
{
  char opt;
  char *lwin;
  
  while((opt=GetSwitch()) != NOSWT)
    switch(opt) {
#ifdef SPHERE
    case 'S':
      sphere=true;
      break;
#endif /* SPHERE */
    case 'f':
      if(NextArgType() != NUMARG) {switcherr('f');}
      freq=GetNumArg();
      break;
    case 'i':
      if(NextArgType() != NUMARG) {switcherr('i');}
      f_min=GetNumArg();
      break;
    case 'u':
      if(NextArgType() != NUMARG) {switcherr('i');}
      f_max=GetNumArg();
      break;
    case 'k':
      if(NextArgType() != NUMARG) {switcherr('k');}
      emphco=GetNumArg();
      break;
    case 'l':
      if(NextArgType() != NUMARG) {switcherr('l');}
      fm_l=GetNumArg();
      break;
    case 'd':
      if(NextArgType() != NUMARG) {switcherr('d');}
      fm_d=GetNumArg();
      break;
    case 'w':
      if(NextArgType() != STRARG) {switcherr('w');}
      lwin=GetStrArg();
      if(strcasecmp(lwin,"none") == 0)
	win=RECTANGLE;
      else if(strcasecmp(lwin,"hamming") == 0)
	win=HAMMING;
      else if(strcasecmp(lwin,"hanning") == 0)
	win=HANNING;
      else if(strcasecmp(lwin,"blackman") == 0)
	win=BLACKMAN;
      else {switcherr('w');}
      break;
    case 'n':
      if(NextArgType() != NUMARG) {switcherr('n');}
      nfilt=(int)GetNumArg();
      break;
    case 'a':
      if(NextArgType() != NUMARG) {switcherr('a');}
      alpha=GetNumArg();
      alphaSet=true;
      break;
    case 'm':
      useMel=true;
      break;
    case 'p':
      if(NextArgType() != NUMARG) {switcherr('p');}
      numceps=(int)GetNumArg();
      break;
    case 'z':
      cms=true;
      break;
    case 'r':
      if(NextArgType() != NUMARG) {switcherr('r');}
      lifter=(int)GetNumArg();
      break;
    case 'e':
      flag |= WITHE;
      break;
    case 's':
      if(NextArgType() != NUMARG) {switcherr('s');}
      escale=GetNumArg();
      break;
    case 'v':
      trace=true;
      break;
    case 'V':
      fprintf(stdout,"\n@%s -- %s\n\n",SPRO_V,cvsid);
      exit(0);
      break;
    case 'h':
      return(false);
      break;
    case '-':
      return(false); break;      
    default:
      gerror(ARGS_ERR,"Unknown switch -- %c",opt);
      return(false);
    }
  return(true);
}

#undef _sfbcep_c_
