/************************************************************/
/*                          slist.c                         */
/*                                                          */
/*                        SPro Sources                      */
/* -------------------------------------------------------- */
/* Guig                                           Apr. 1997 */
/************************************************************/
/*
 * CVS log:
 *
 * $Author: gravier $
 * $Date: 1999/12/02 16:54:34 $
 * $Revision: 1.3 $
 *
 */
/*
 * List content of a data file to screen
 * Possible convertions are allowed as in scopy
 *
 * -----
 *
 * Input: SPro data file
 * Output:
 */
#define _slist_c_
#ifndef lint
static char *cvsid = "$Id: slist.c,v 1.3 1999/12/02 16:54:34 gravier Exp $";
#endif

#include <stdio.h>
#include <stdlib.h>

#include <gutil.h>
#include <gerror.h>
#include <errcode.h>
#include <spro.h>

#define LINELEN    6       /* Number of parameters per line */

/* ---------------------------- */
/* ----- void usage(void) ----- */
/* ---------------------------- */
void usage(void)
{
  fprintf(stderr,"\n@%s\n",SPRO_V);
  fprintf(stderr,"\nUsage:\n"); 
  fprintf(stderr,"    %s [options] datafile\n",prog);
  fprintf(stderr,"\n");
  fprintf(stderr,"Synopsis:\n");
  fprintf(stderr,"    List content of a data file\n");
  fprintf(stderr,"\n");
  fprintf(stderr,"Options:\n");
  fprintf(stderr,"  --info, -i                show header only (off)\n");
  fprintf(stderr,"  --start, -s num           starting at sample number (1)\n");
  fprintf(stderr,"  --end, -e num             ending at sample number (last)\n");
  fprintf(stderr,"  --convert, -C str         convertion string - ADNZ (none)\n");
  fprintf(stderr,"  --version, -V             print version number\n");
  fprintf(stderr,"\n");
}

/* ----------------------------- */
/* ----- Long option array ----- */
/* ----------------------------- */
swtab_t longopts[]={
  {"info",'i'},
  {"start",'s'},
  {"end",'e'},
  {"convert",'C'},
  {"version",'V'},
  {"help",'h'},
  {NULL,0x00}
};
/* ---------------------------------------------- */
/* ----- Global variables set by ReadArgs() ----- */
/* ---------------------------------------------- */
bool_t showdata=true;  /* Set to false with -i option          */
unsigned long st=1;    /* Start at sample st (default: first)  */
unsigned long en=0;    /* Start at sample en (default: last)   */
flag_t cvtflag=(flag_t)0;

/* --------------------------------- */
/* ----- int main(int,char **) ----- */
/* --------------------------------- */
int main(int argc,char **argv)
{
  char *ifile;           /* Input filename                       */
  FILE *f;               /* I/O file descriptor                  */
  data_t *data;          /* Input data structure                 */
  err_t status;
  bool_t ReadArgs(void);
  bool_t CheckArgs(void);
  void ShowHeader(data_t *);
  void ShowData(data_t *);
  
  /***** Process input arguments *****/
  LongArgInit(argc,argv,longopts);
  if(!ReadArgs()) {
    usage();
    return(ARGS_ERR);
  }
  if(!CheckArgs())
    return(ARGS_ERR);
  if(NextArgType() != STRARG) {
    gerror(ARGS_ERR,"%s: no input file specified",prog);
    usage();
    return(ARGS_ERR);
  }
  ifile=GetStrArg();

  /***** Load data from input file *****/
  if((f=fopen(ifile,"rb")) == NULL) {
    gerror(FOPEN_ERR,"%s: cannot open file %s",prog,ifile);
    return(FOPEN_ERR);
  }
  if((status=LoadSPData(&data,cvtflag,f)) != OK) {
    gerror(status,"%s: cannot load data from file %s",prog,ifile);
    fclose(f);
    return(status);
  }
  fclose(f);

  /***** Show on screen what was asked *****/
  ShowHeader(data);
  if(showdata)
    ShowData(data);

  /***** House keeping *****/
  FreeSPData(data);
  DONE;
}

/**************************************/
/********** Display routines **********/
/**************************************/

/* ------------------------------------- */
/* ----- void ShowHeader(data_t *) ----- */
/* ------------------------------------- */
/*
 * Display data header
 */
void ShowHeader(data_t *data)
{
  void ShowFlag(flag_t);

  fprintf(stdout,"--------------------------------\n");
  fprintf(stdout,"Datakind: %s\n",kind2str(data->kind));
  fprintf(stdout,"Number of samples: %lu\n",data->n);
  fprintf(stdout,"Base sample size: %lu\n",data->dim);
  fprintf(stdout,"Qualifiers: ");
  ShowFlag(data->flag);
  fprintf(stdout,"\n--------------------------------\n");
}

/* --------------------------------- */
/* ----- void ShowFlag(flag_t) ----- */
/* --------------------------------- */
/*
 * Display qualifier flags
 */
void ShowFlag(flag_t flag)
{
  if(flag & WITHE) fprintf(stdout," E");
  if(flag & WITHD) fprintf(stdout," D");
  if(flag & WITHA) fprintf(stdout," A");
  if(flag & WITHN) fprintf(stdout," N");
  if(flag & WITHZ) fprintf(stdout," Z");
}

/* ----------------------------------- */
/* ----- void ShowData(data_t *) ----- */
/* ----------------------------------- */
/*
 * Display samples between st and en.
 * Static coeffs are preceded by s, delta coeffs by d
 * andd delta-delta by a.
 *
 * The constant LINELEN defined at the beginning of the code
 * controls the number of samples displayed per lines.
 */
void ShowData(data_t *data)
{
  int nelem,i,j,p;

  nelem=DataVecSize(data);
  if(en == 0)
    en=data->n;
  if(en > data->n || st > data->n) {
    gerror(RANGE_ERR,"%s: invalid start and end values (st= %ul, en= %ul, nsamples= %lu)",
	   prog,st,en,data->n);
    return;
  }

  for(i=st;i<=en;i++) {
    fprintf(stdout,"\nSample %d:",i);
    p=0;
    for(j=0;j<data->dim;j++) {
      if(j%LINELEN == 0)
	fprintf(stdout,"\n    ");
      fprintf(stdout," s%02d: %-8.4f",j,data->s[(i-1)*nelem+j]);
    }
    if(data->flag & WITHE && !(data->flag & WITHN)) {
      fprintf(stdout,"\n     sE: %-8.4f",data->s[(i-1)*nelem+data->dim]);
      p=data->dim+1;
    }
    else
      p=data->dim;
    if(data->flag & WITHD) {
      for(j=0;j<data->dim;j++,p++) {
	if(j%LINELEN == 0)
	  fprintf(stdout,"\n    ");
	fprintf(stdout," d%02d: %-8.4f",j,data->s[(i-1)*nelem+p]);
      }
      if(data->flag & WITHE) {
	fprintf(stdout,"\n     dE: %-8.4f",data->s[(i-1)*nelem+p]);
	p++;
      }
    }
	
    if(data->flag & WITHA) {
      for(j=0;j<data->dim;j++,p++) {
	if(j%LINELEN == 0)
	  fprintf(stdout,"\n    ");
	fprintf(stdout," a%02d: %-8.4f",j,data->s[(i-1)*nelem+p]);
      }
      if(data->flag & WITHE)
	fprintf(stdout,"\n     aE: %-8.4f",data->s[(i-1)*nelem+p]);
    }
  }
  fprintf(stdout,"\n\n");
}

/**************************************************/
/********** Argument processing routines **********/
/**************************************************/
/* ---------------------------------- */
/* ----- bool_t CheckArgs(void) ----- */
/* ---------------------------------- */
/*
 * Check that arguments are valid
 */
bool_t CheckArgs(void)
{
  if(en && st > en) {
    gerror(ARGS_ERR,"%s: start sample greater than end sample (st=%d, en=%d)",
	   prog,st,en);
    return(false);
  }
  if(st < 1 || (en && en < 1)) {
    gerror(ARGS_ERR,"%s: invalid start and/or end sample (st=%d, en=%d)",
	   prog,st,en);
    return(false);
  }
  return(true);
}

/* --------------------------------- */
/* ----- bool_t ReadArgs(void) ----- */
/* --------------------------------- */
/*
 *  --info, -i                show header only (off)\n");
 *  --start, -s num           starting at sample number (1)\n");
 *  --end, -e num             ending at sample number (last)\n");
 *  --convert, -C str         convertion string - ADNZ (none)\n");
 *  --version, -V             print version number\n");
 *  --help
 */
bool_t ReadArgs(void)
{
  char opt;

  while((opt=GetSwitch()) != NOSWT)
    switch(opt) {
    case 'i':
      showdata=false;
      break;
    case 's':
      if(NextArgType() != NUMARG) {switcherr('s');}
      st=(unsigned long)GetNumArg();
      break;
    case 'e':
      if(NextArgType() != NUMARG) {switcherr('e');}
      en=(unsigned long)GetNumArg();
      break;
    case 'C':
      if(NextArgType() != STRARG) {switcherr('C');}
      cvtflag=ConvStr2Flag(GetStrArg());
      break;
    case 'V':
      fprintf(stdout,"\n@%s -- %s\n\n",SPRO_V,cvsid);
      exit(0);
      break;
    case 'h':
      return(false);
      break;
    case '-':
      return(false); break;      
    default:
      gerror(ARGS_ERR,"Unknown option switch -- %c\n",opt);
      return(false);
    }
  return(true);
}

#undef _slist_c_
     
