/************************************************************/
/*                         slpc.c                           */
/*                                                          */
/*                      SPro Sources                        */
/* -------------------------------------------------------- */
/* Guig                                           Apr. 1997 */
/************************************************************/
/*
 * CVS modification log:
 *
 * $Author: gravier $
 * $Date: 1999/12/03 13:20:40 $
 * $Revision: 1.4 $
 *
 */
/*
 * This tool was formerly siglpc.c in SPro v2.0
 * It acts the same and is designed to perform LPC analysis
 * of a signal.
 *
 * It can output LPCOEFF, PARCOR or LAR data files.
 * Energy (if asked) is actually the prediction error, that is
 * sigma if the AR filter is sigma/A(z).
 *
 * -----
 *
 * Input:  raw, linear 16 bits/sample signal file, or
 *         SPHERE file
 * Output: SPro LPCOEFF, PARCOR or LAR data file
 */

#define _slpc_c_
#ifndef lint
static char *cvsid = "$Id: slpc.c,v 1.4 1999/12/03 13:20:40 gravier Exp $";
#endif

#include <stdio.h>
#include <stdlib.h>

#include <gutil.h>
#include <gerror.h>
#include <errcode.h>
#include <spro.h>

/* ---------------------------- */
/* ----- void usage(void) ----- */
/* ---------------------------- */
void usage(void)
{
  fprintf(stderr,"\n@%s\n",SPRO_V);
  fprintf(stderr,"\nUsage:\n");
  fprintf(stderr,"    %s [options] sigfile datafile\n",prog);
  fprintf(stderr,"\n");
  fprintf(stderr,"Synopsis:\n");
  fprintf(stderr,"    Perform variable resolution AR modelling of the input signal.\n");
  fprintf(stderr,"\n");
  fprintf(stderr,"Options:\n");
#ifdef SPHERE
  fprintf(stderr,"  --sphere, -S              Sphere signal file format (pcm)\n");
#endif /* SPHERE */
  fprintf(stderr,"  --sample-rate, -f num     signal sample rate (8000 Hz)\n");
  fprintf(stderr,"  --pre-emphasis, -e num    pre-emphasis coefficient (0.95)\n");
  fprintf(stderr,"  --length, -l num          frame length in ms (20)\n");
  fprintf(stderr,"  --decay, -d num           frame decay in ms (10)\n");
  fprintf(stderr,"  --window, -w str          weighting window (hamming)\n");
  fprintf(stderr,"  --order, -p num           prediction order (16)\n");
  fprintf(stderr,"  --alpha, -a num           frequency warping parameter (0.0)\n");
  fprintf(stderr,"  --parcor, -k              output reflexion coefficients (LPC)\n");
  fprintf(stderr,"  --lar, -g                 output log area ratios (LPC)\n");
  fprintf(stderr,"  --energy, -E              add log-energy to ouput file\n");  
  fprintf(stderr,"  --scale-energy, -s num    scale energy (off)\n");
  fprintf(stderr,"  --verbose, -v             verbose mode\n");
  fprintf(stderr,"  --version, -V             print version number\n");
  fprintf(stderr,"\n");
}

/* ----------------------------- */
/* ----- Long option array ----- */
/* ----------------------------- */
swtab_t longopts[]={
#ifdef SPHERE
  {"sphere",'S'},
#endif /* SPHERE */
  {"sample-rate",'f'},
  {"pre-emphasis",'e'},
  {"length",'l'},
  {"decay",'d'},
  {"window",'w'},
  {"order",'p'},
  {"alpha",'a'},
  {"parcor",'k'},
  {"lar",'g'},
  {"energy",'e'},
  {"scale-energy",'s'},
  {"verbose",'v'},
  {"version",'V'},
  {"help",'h'},
  {NULL,0x00}
};

/* ---------------------------------------------- */
/* ----- Global variables set by ReadArgs() ----- */
/* ---------------------------------------------- */
bool_t trace=false;     /* Set to true with -v option           */
float freq=8000.0;      /* Sample rate (default: 8 KHz)         */
int p=16;               /* Prediction order (default: 12)       */
float fm_l=20.0;        /* Frame length (default: 20 ms)        */
float fm_d=10.0;        /* Frame decay (default: 10 ms)         */
float alpha=0.0;        /* frequency deformation parameter      */
float emphco=0.95;      /* pre-emphasis coefficient             */
win_t win=HAMMING;      /* weighting window                     */
flag_t flag=0;          /* Qualifier flag                       */
datakind_t kind=LPCOEFF;/* What is the output?                  */
bool_t gset=false;
bool_t kset=false;
float escale=0.0;       /* Energy scaling factor                */
#ifdef SPHERE
bool_t sphere=false;   /* Set to true with -S flag             */
#endif /* SPHERE */

/* --------------------------------- */
/* ----- int main(int,char **) ----- */
/* --------------------------------- */
int main(int argc,char **argv)
{
  char *ifile;            /* input signal filename                 */
  char *ofile;            /* output file descriptor                */
  FILE *f;                /* file descriptor                       */
  sig_t *sig;             /* input signal                          */
  data_t *data;           /* Output data structure                 */
  err_t status;           /* Return status from LPAnalysis()       */
  int nframe;             /* Number of frames in sig               */
  bool_t ReadArgs(void);  /* Read and parse option switches        */
  bool_t CheckArgs(void); /* Check option validity                 */
  err_t LPAnalysis(sig_t *,data_t *);
  
  /***** Process input arguments ... *****/
  LongArgInit(argc,argv,longopts);
  if(!ReadArgs()) {
    usage();
    return(ARGS_ERR);
  }
  if(!CheckArgs())
    return(ARGS_ERR);
  if(NextArgType() != STRARG) {
    gerror(ARGS_ERR,"%s: no input signal file specified",prog);
    usage();
    return(ARGS_ERR);
  }
  ifile=GetStrArg();
  if(NextArgType() != STRARG) {
    gerror(ARGS_ERR,"%s: no output file specified",prog);
    usage();
    return(ARGS_ERR);
  }
  ofile=GetStrArg();

  /***** Show what was asked to do *****/
  if(trace) {
    fprintf(stdout,"\n%s (%s): %s --> %s\n\n",prog,SPRO_V,ifile,ofile);
    fflush(stdout);
  }

  /***** Load input signal *****/
#ifdef SPHERE
  if(sphere) {
    if((status=ReadSphereSig(&sig,&freq,ifile)) != OK) {
      gerror(status,"%s: cannot load signal from file %s",prog,ifile);
      return(status);
    }
  }
  else {
#endif /* SPHERE */
    if((f=fopen(ifile,"rb")) == NULL) {
      gerror(FOPEN_ERR,"%s: cannot open input file %s",prog,ifile);
      return(FOPEN_ERR);
    }
    if((status=ReadSig(&sig,fsize(f)/2,f)) != OK) {
      gerror(status,"%s: cannot load signal from file %s",prog,ifile);
      fclose(f);
      return(status);
    }
    fclose(f);
#ifdef SPHERE
  }
#endif /* SPHERE */

  /***** Run pre-emphasis *****/
  (void)SigEmph(sig,emphco);

  /***** Get memory for the output structure *****/
  nframe=1+(sig->n-(int)(fm_l*freq/1000.0))/(int)(fm_d*freq/1000.0);
  if((status=CreateSPData(&data,kind,p,nframe,flag)) != OK) {
    gerror(status,"%s: cannot create output data structure",prog);
    FreeSig(sig);
    return(status);
  }

  /***** Run LPC analysis() *****/
  if((status=LPAnalysis(sig,data)) != OK) {
    gerror(status,"%s: fatal error",prog);
    FreeSig(sig);
    FreeSPData(data);
    return(status);
  }
  if(escale != 0.0)
    ScaleEnergy(data,escale);
  
  /***** Save data to output file *****/
  if((f=fopen(ofile,"wb")) == NULL) {
    gerror(FOPEN_ERR,"%s: cannot open output file %s",prog,ofile);
    FreeSig(sig);
    FreeSPData(data);
    return(FOPEN_ERR);
  }
  if((status=WriteSPData(data,f)) != OK) {
    gerror(status,"%s: cannot dump data to output file %s\n",prog,ofile);
    FreeSig(sig);
    FreeSPData(data);
    return(DUMP_ERR);
  }
  fclose(f);

  FreeSig(sig);
  FreeSPData(data);

  DONE;
}

/* ---------------------------------------------- */
/* ----- err_t LPAnalysis(sig_t *,data_t *) ----- */
/* ---------------------------------------------- */
/*
 * Here is the main encoding procedure which acts as follows:
 *  - get a frame of signal
 *  - pre-process it
 *  - Apply linear prediction
 *  - write coefficients to data structure
 *  - write log energy (if any) to data structure
 */
err_t LPAnalysis(sig_t *sig,data_t *data)
{
  int i,j;
  int nelem;
  float sigma;
  int nframe=0;
  sig_t *frame;
  vector_t *a;
  err_t status;

  /***** Get necessary memory *****/
  if((status=CreateSig(&frame,(unsigned long)(fm_l*freq/1000.0))) != OK)
    return(status);
  if((status=CreateVector(&a,p)) != OK) {
    FreeSig(frame);
    return(status);
  }
  nelem=DataVecSize(data);

  /***** Process every frame *****/
  for(i=1;i<=data->n;i++) {
    /*
     * Get and pre-process frame
     */
    if((status=SigGetFrame(sig,i,frame->n,(int)(fm_d*freq/1000.0),frame)) != OK) {
      gerror(status,"%s: error reading frame %d",prog,i);
      return(status);
    }
    SigWeight(frame,win);
    /*
     * Run frame level LP analysis
     */
    if((status=ARGModel(frame,alpha,p,kind,a,&sigma)) != OK) {
      gerror(status,"%s: cannot apply LP analysis to frame %d",prog,i);
      FreeSig(frame); FreeVector(a);
      return(status);
    }
    /*
     * Copy parameter vector to data
     */
    nframe++;
    for(j=0;j<data->dim;j++)
      data->s[(nframe-1)*nelem+j]=a->data[j];
    if(data->flag & WITHE)
      data->s[nframe*nelem-1]=sigma < ENERGY_FLOOR ?
	(param_t)(2.0*log(ENERGY_FLOOR)) : (param_t)(2.0*log(sigma));
  }
  data->n=nframe;

  /**** Give memory back and say OK *****/
  FreeSig(frame);
  FreeVector(a);

  DONE;
}

/* ---------------------------------- */
/* ----- bool_t CheckArgs(void) ----- */
/* ---------------------------------- */
bool_t CheckArgs(void)
{
  if(p < 1 || p > MAX_LPCORDER) {
    gerror(ARGS_ERR,"%s: invalid LPC analysis order %d (min= 1, max= %d)",
	    prog,p,MAX_LPCORDER);
    return(false);
  }
  if(kset && gset) {
    gerror(ARGS_ERR,"%s: -k and -g used together",
	   prog);
    return(false);
  }
  return(true);
}

/* --------------------------------- */
/* ----- bool_t ReadArgs(void) ----- */
/* --------------------------------- */
/*
 *  --sphere, -S              Sphere signal file format (pcm)\n");
 *  --sample-rate, -f num     signal sample rate (8000 Hz)\n");
 *  --pre-emphasis, -e num    pre-emphasis coefficient (0.95)\n");
 *  --length, -l num          frame length in ms (20)\n");
 *  --decay, -d num           frame decay in ms (10)\n");
 *  --window, -w str          weighting window (hamming)\n");
 *  --order, -p num           prediction order (16)\n");
 *  --alpha, -a num           frequency warping parameter (0.0)\n");
 *  --parcor, -k              output reflexion coefficients (LPC)\n");
 *  --lar, -g                 output log area ratios (LPC)\n");
 *  --energy, -E              add log-energy to ouput file\n");  
 *  --scale-energy, -s num    scale energy (off)\n");
 *  --verbose, -v             verbose mode\n");
 *  --version, -V             print version number\n");
 *  --help
 */
bool_t ReadArgs(void)
{
  char opt;
  char *lwin;
  
  while((opt=GetSwitch()) != NOSWT)
    switch(opt) {
#ifdef SPHERE
    case 'S':
      sphere=true;
      break;
#endif /* SPHERE */
    case 'f':
      if(NextArgType() != NUMARG) {switcherr('f');}
      freq=GetNumArg();
      break;
    case 'e':
      if(NextArgType() != NUMARG) {switcherr('e');}
      emphco=GetNumArg();
      break;
    case 'l':
      if(NextArgType() != NUMARG) {switcherr('l');}
      fm_l=GetNumArg();
      break;
    case 'd':
      if(NextArgType() != NUMARG) {switcherr('d');}
      fm_d=GetNumArg();
      break;
    case 'w':
      if(NextArgType() != STRARG) {switcherr('w');}
      lwin=GetStrArg();
      if(strcasecmp(lwin,"none") == 0)
	win=RECTANGLE;
      else if(strcasecmp(lwin,"hamming") == 0)
	win=HAMMING;
      else if(strcasecmp(lwin,"hanning") == 0)
	win=HANNING;
      else if(strcasecmp(lwin,"blackman") == 0)
	win=BLACKMAN;
      else {switcherr('w');}
      break;
    case 'p':
      if(NextArgType() != NUMARG) {switcherr('p');}
      p=(int)GetNumArg();
      break;
    case 'a':
      if(NextArgType() != NUMARG) {switcherr('a');}
      alpha=GetNumArg();
      break;
    case 'k':
      kind=PARCOR;
      kset=true;
      break;
    case 'g':
      kind=LAR;
      gset=true;
      break;
    case 'E':
      flag |= WITHE;
      break;
    case 's':
      if(NextArgType() != STRARG) {switcherr('s');}
      escale=GetNumArg();
      break;
    case 'v':
      trace=true;
      break;
    case 'V':
      fprintf(stdout,"\n@%s -- %s\n\n",SPRO_V,cvsid);
      exit(0);
      break;      
    case 'h':
      return(false);
      break;
    case '-':
      return(false); break;      
    default:
      gerror(ARGS_ERR,"Unknown switch -- %c",opt);
      return(false);
    }
  return(true);
}

#undef _slpc_c_
