/************************************************************/
/*                          spro.h                          */
/*                                                          */
/*                       SPro Library                       */
/* -------------------------------------------------------- */
/* Guig                                           Apr. 1997 */
/************************************************************/
/*
 * CVS log:
 *
 * $Author: gravier $
 * $Date: 1999/12/03 13:23:43 $
 * $Revision: 1.6 $
 *
 */
/*
 * I consider that a waveform is NOT a data.
 * sig_t is provided for waveforms (and for compatibilty with
 * older versions of SPro) while data_t is provided for speech
 * parameters, such as LPC or LPCEPSTRA, and energy.
 */

#ifndef _spro_h_
#define _spro_h_

#include <limits.h>                   /* Standard C headers                    */
#include <math.h>

#include <gutil.h>                    /* gglib headers                         */
#include <gerror.h>
#include <errcode.h>
#include <sproerr.h>

#define SPRO_V              "SPro v3.2"

/*****
 * Constant definitions
******/
#define MAX_FRAMESIZE       40         /* Maximum length for a frame           */
#define MAX_FREQUENCY       48000.0    /* Maximum sampling frequency           */
#define MAX_SAMPLESINFRAME  1920       /* 48 KHz * 40 ms = 1920 samples        */
#define MAX_CEPSTRUM        20         /* Maximum number of cepstral coeffs    */
#define MAX_LPCORDER        30         /* maximum order for LPC analysis       */
#define MAX_BANK_FILTER     24         /* maximum nb. of filters in a bank     */
#define MIN_BANK_FILTER     2          /* miumium nb. of filters in a bank     */
#define FBANK_FFT_SIZE      512        /* Length of FFT for filter banks       */
#define MAX_FFT_SIZE        512        /* Maximum admissible FFT length        */
#define FBANK_OUT_FLOOR     1.0        /* Energy below this floor is floored   */
#define ENERGY_FLOOR        1.0        /* Energy below this floor is floored   */
#define MAX_QSMP            SHRT_MAX   /* Max. for a quantized waveform sample */
#define MIN_QSMP            SHRT_MIN   /* Min. for a quantized waveform sample */
#define MAX_STATIC_SIZE     24         /* Max. size for static coefficients    */

#define WITHE               0x01       /* Data is with log energy              */
#define WITHZ               0x02       /* Data as mean suppressed (static only)*/
#define WITHN               0x04       /* Absolute log energy suppressed       */
#define WITHD               0x08       /* Data as delta coefficients           */
#define WITHA               0xa0       /* Data contains delta-delta coeffs     */

#define round(x) (int)ceil(x)-x<x-(int)floor(x)?(int)ceil(x):(int)floor(x)

/*****
 * Type definitions
*****/
     /***** Data kind, type and structure *****/

typedef enum {
  OTHER,                             /* Anything else I didn't think about        */
  FBANK,                             /* Filter bank output                        */
  FBCEPSTRA,                         /* Cepstra from filter bank analysis         */
  LPCEPSTRA,                         /* Cepstra from LPC analysis                 */
  LPCOEFF,                           /* Linear prediction coefficients            */
  PARCOR,                            /* Partial correlation coefficients          */
  LAR                                /* Log area ratios                           */
} datakind_t;                        /* Definition of sample kinds                */

typedef float sample_t;              /* Definition of a sample type               */
typedef float param_t;               /* Definition of a parameter type            */
typedef unsigned int flag_t;         /* Flag for qualifiers                       */

typedef struct {
  datakind_t kind;                   /* Sample kind                               */
  unsigned long dim;                 /* Vector dimension                          */
  unsigned long n;                   /* Number of samples                         */
  flag_t flag;                       /* Qualifiers (4 bytes for compatibility)    */
  param_t *s;                        /* Pointer to the samples                    */
} data_t;                            /* Structure for all kind of data            */

typedef struct {
  unsigned long n;                   /* Number of samples                         */
  sample_t *s;                       /* Array of samples                          */
} sig_t;                             /* A real signal (i.e. waveform)             */

     /***** spro_math type definitions *****/

typedef struct {
  unsigned long n;                   /* Number of elements                        */
  param_t *data;                     /* Vector elements                           */
} vector_t;

     /***** spro_global type definitions *****/

typedef enum {RECTANGLE,HAMMING,HANNING,BLACKMAN} win_t;
typedef short int qsmp_t;       /* Quantized sample type (for sig_t only)  */

     /***** spro_io function prototypes *****/
/* Create s signal structure */
err_t CreateSig(
  sig_t **,                     /* Pointer adress for the signal           */
  unsigned long                 /* Number of samples                       */
);

/* Free a signal */
void FreeSig(
  sig_t *                       /* Pointer to signal                       */
);

/* Read a signal */
err_t ReadSig(
  sig_t **,                     /* Pointer adress for the signal           */
  unsigned long,                /* Number of samples                       */
  FILE *                        /* Input stream                            */
);

#ifdef SPHERE
/* Read a signal from a sphere file */
err_t ReadSphereSig(
  sig_t **,                     /* Pointer adress for the signal           */
  float *,                      /* Sample rate if any in the sphere header */
  char *                        /* Input filename                          */
);
#endif

/* Write a signal */
err_t WriteSig(
  sig_t *,                      /* Pointer to the signal                    */
  FILE *                        /* Output stream                            */
);

/* Create a data stucture */
err_t CreateSPData(
  data_t **,                    /* Pointer adress for the data              */
  datakind_t,                   /* Kind of samples to store                 */
  unsigned long,                /* Sample dimension (without energy)        */
  unsigned long,                /* Number of samples                        */
  flag_t                        /* Qualifiers                               */
);

/* Free a data structure */
void FreeSPData(
  data_t *                      /* The structure to be freed                */
);

/* Read data from an input stream */
err_t ReadSPData(
  data_t **,                    /* Pointer adress for the data              */
  FILE *                        /* Input stream                             */
);

/* Write data to an output stream */
err_t WriteSPData(
  data_t *,                     /* Pointer to the data                      */
  FILE *                        /* Output stream                            */
);

/* Load data (e.g. read and convert if necessary) 
   replace ReadSPData whenever a convertion may be necessary */
err_t LoadSPData(
  data_t **,                    /* Pointer adress for the data              */
  flag_t,                       /* Convertion flag                          */
  FILE *                        /* Input stream                             */
);

/* Allocate memory for a vector */
err_t CreateVector(
  vector_t **,                  /* Pointer adress for the vector            */
  unsigned long                 /* Vector dimension                         */
);

/* Free memory */
void FreeVector(
  vector_t *                    /* Pointer to the vector                    */
);

/* Write data to HTK format (with USER type) */
err_t HTKWrite(
  data_t *,                     /* Pointer to the structure to write        */
  float,                        /* Sample period (in ms)                    */
  int,                          /* HTK version (should be 1 or 2)           */
  FILE *                        /* Output stream                            */
);

     /***** spro_global function prototypes *****/

/* Convertions of enum types to a string */
char *kind2str(datakind_t);
char *win2str(win_t);

/* Remove static coefficients mean */
void DataZeroMean(
  data_t *                      /* pointer to a data structure              */
);

/* Return full feature vector size (with E and/or D and/or A) */
unsigned long DataVecSize(
  data_t *                      /* Pointer to data structure                */
);

/* Return a pointer to the requested data */
param_t *DataGetVec(
  data_t *,                     /* Pointer to data structure                */
  int,                          /* Vector number starting at 0              */
  unsigned long                 /* Vector size                              */
);

/* Cepstral liftering */
void CepLifter(
  vector_t *,                   /* Vector of cepstral coefficients          */
  int                           /* Liftering value                          */
);

/* Energy normalization and scaling */
void ScaleEnergy(
  data_t *,                     /* Pointer to data structure                */
  float                         /* Scaling factor                           */
);
/* Get a frame in a longer signal */
err_t SigGetFrame(
  sig_t *,                      /* Input signal                             */
  int,                          /* Frame number (starting with frame 1)     */
  int,                          /* Frame length (in samples)                */
  int,                          /* Frame decay (in samples)                 */
  sig_t *                       /* frame signal (must be allocated before)  */
);

/* Center a signal (return mean value) */
float SigCenter(
  sig_t *,                      /* Input (and output) signal                */
  bool_t                        /* Action flag                              */
);

/* Normalize a signal (return energy) */
float SigNorm(
  sig_t *,                      /* Input (and output) signal                */
  bool_t                        /* Action flag                              */
);

/* Weighting of a signal (should actually be a frame) */
err_t SigWeight(
  sig_t *,                      /* Input (and output) signal                */
  win_t                         /* Weighting window                         */
);

/* pre-emphasis and de-emphasis */
err_t SigEmph(
  sig_t *,                      /* Input (and output) signal                */
  float                         /* coefficient, >0 for emphasis or <0       */
);

/* Divide sample rate by 2 as many times as requested (return down sampled signal) */
err_t SigDownSample(
  sig_t **,                     /* Pointer adress for the returned signal   */
  sig_t *,                      /* Input signal                             */
  int                           /* Number of decimation                     */
);

     /***** spro_conv function prototypes *****/

/* Add/Remove qualifiers */
err_t ConvertData(
  data_t **,                    /* Pointer adress for the returned data     */
  data_t *,                     /* Original data to convert                 */
  flag_t                        /* Converted data type                      */
);

/* Convert a DANZ string into a flag */
flag_t ConvStr2Flag(
  char *                        /* DANZ string                              */
);

#ifdef _spro_conv_c_
/*
 * Private procedure
 *
 */
err_t CheckConv(datakind_t,flag_t,datakind_t,flag_t);
#endif /* _spro_conv_c_ */

     /***** spro_lpc function prototypes *****/

/* Generalized linear prediction (variable spectral analysis) */
err_t ARGModel(
  sig_t *,                      /* Input signal                             */
  float,                        /* Spectral deformation parameter (alpha)   */
  int,                          /* Analysis order                           */
  datakind_t,                   /* What kind of coefficients?               */
  vector_t *,                   /* Vector of parameter (must be allocated)  */
  float *                       /* Prediction error                         */
);

/* LP inverse filter [A(z)/sigma] */
err_t SigARInverse(
  sig_t *,                      /* Allocated residual signal                */
  vector_t *,                   /* Coefficient vector (and order)           */
  datakind_t,                   /* What kind of coefficients?               */
  float,                        /* sigma                                    */
  int,                          /* Start sample (from 0)                    */
  int,                          /* End sample                               */
  sig_t *                       /* Input signal                             */
);

/* LP direct filter [sigma/A(z)] */
err_t SigARFilter(
  sig_t *,                      /* Allocated filtered signal                */
  vector_t *,                   /* Coefficient vector (and order)           */
  datakind_t,                   /* What kind of coefficients?               */
  float,                        /* sigma                                    */
  int,                          /* Start sample (from 0)                    */
  int,                          /* End sample (included)                    */
  sig_t *                       /* Input (residual/excitation) signal       */
);

/* Coefficient conversion */
err_t ARConvert(
  vector_t *,                   /* Input vector of parameters               */
  datakind_t,                   /* Source kind                              */
  datakind_t,                   /* Target kind                              */
  vector_t *                    /* Output vector (must be allocated before) */
);

/* LPCOEFF ---> LPCEPSTRA */
err_t LPCepstrum(
  vector_t *,                   /* Input vector of LPCOEFF parameters       */
  int,                          /* Number of cepstral coefficients          */
  vector_t *                    /* Output vector (must be allocated before) */
);

#ifdef _spro_lpc_c_
/*
 * Private procedures
 * 
 */
/*
 * Correlation between the output of dephasing filters
 */
err_t ARGCorrel(sig_t *,float,float *,int);
/*
 * Auto-correlation
 */
void AutoCorrel(sig_t *,float *,int);
/* 
 * LAR <---> PARCOR <---> LPCOEFF conversions
 */
void a2k(param_t *,unsigned long,param_t *);
void k2a(param_t *,unsigned long,param_t *);
void g2k(param_t *,unsigned long,param_t *);
void k2g(param_t *,unsigned long,param_t *);
#endif /* _spro_lpc_c_ */

          /***** spro_fft function prototypes *****/

/* Perform FFT on a signal (interface to Nielsen's mixfft) */
err_t FFTransform(
  sig_t *,                      /* Input signal                                               */
  unsigned long,                /* length of the transformation (may not be power of two)     */
  vector_t *,                   /* Output vector of modules (npts/2 allocated before) or NULL */
  vector_t *                    /* Output vector of phases  (npts/2 allocated before) or NULL */
);

/* Log Filter bank output on variable frequency scales */
err_t LogFilterBank(
  sig_t *,                      /* Input signal                             */
  float,                        /* Spectral deformation parameter (alpha)   */
  int,                          /* Number of filters                        */
  float,                        /* min normalized frequency (in [0,1/2])    */
  float,                        /* max normalized frequency (in [0,1/2])    */
  vector_t *                    /* Vector of parameter (must be allocated)  */
);

/* Log Filter bank output on Mel scale */
err_t LogMelFilterBank(
  sig_t *,                      /* Input signal                             */
  float,                        /* Sample rate of input signal              */
  int,                          /* Number of filters                        */
  float,                        /* min frequency (in Hz)                    */
  float,                        /* max frequency (in Hz)                    */
  vector_t *                    /* Vector of parameter (must be allocated)  */
);

err_t DCTransform(
  vector_t *,                   /* Input vector of parameters                      */
  int,                          /* Number of output coefficients                   */
  vector_t *                    /* Output vector of parameters (must be allocated) */
);

#ifdef _spro_fft_c_
/*
 * Private procedures
 *
 */
void fft(float *,unsigned long);
void realfft(float *,unsigned long);
/*
 * Apply a triangular filter bank to a module vector, knowing the
 * indexes of each cut-off frequencies.
 */
void LogTriangularFBank(vector_t *,int,int *,vector_t *);
void TriangularFBank(vector_t *,int,int *,vector_t *);
/*
 * Apply omega'=Theta(omega) (inverse) transformation to a pulsation
 */
float Theta(float,float);
float ThetaInv(float,float);
/*
 * Apply Mel (inverse) transformation to a given frequency
 */
float MelInv(float);
float Mel(float);
#endif /* _spro_fft_c_ */

#endif /* _spro_h_ */
