/************************************************************/
/*                      spro_conv.c                         */
/*                                                          */
/*                       SPro v3.0                          */
/* -------------------------------------------------------- */
/* Guig                                           Apr. 1997 */
/************************************************************/
/*
 * CVS log:
 *
 * $Author: gravier $
 * $Date: 1999/08/19 15:06:50 $
 * $Revision: 1.4 $
 *
 */
/*
 * Convertion routine.
 *
 * Perform convertion of data to bring them in accordance
 * with flag. This is mainly for purpose of adding delta
 * and delta-delta on the fly, as done in HTK (I Hate HTK).
 * At least, it saves disk space ...
 *
 * It is important to note that the base kind is not changed.
 * What is allowed for the moment is to:
 *   - remove mean of static coefficients
 *   - add delta
 *   - add acc
 *   - suppress static log energy
 *
 * NOTE: This is not the complete version as there are many 
 * things I still don't care about!
 *
 * If replace is set to true, then the original data structure is 
 * freed (therefore you can do data=ConvertData(data,flag,true) and
 * the data pointer is replaced). When replace is false, the original
 * data structure is not freed. 
 *
 * The action looks somewhat like this:
 *
 *  for each frame
 *    copy static coefficients
 *    copy energy if necessary
 *    remove static ceoff mean if necessary
 *    compute delta from original data if necessary
 *    compute acc from previously computed delta
 *    
 */

#define _spro_conv_c_

#include <spro.h>

/* -------------------------------------------------------- */
/* ----- err_t ConvertData(data_t **,data_t *,flag_t) ----- */
/* -------------------------------------------------------- */
err_t ConvertData(data_t **tgtdat,data_t *srcdat,flag_t flag)
{
  int i,j,k,p;
  unsigned long srcl,tgtl;
  param_t *srcs,*tgts;
  param_t *p1,*p2,*n1,*n2; /* previous t-1 and t-2 and next t+1 and t+2 */
  err_t status;

  /*
   * Check that there is something to do ...
   */
  if(flag == srcdat->flag) {
    gerror(NULL_CONV_ERR,"ConvertData(): no convertion to do (target == source)");
    *tgtdat=srcdat;
    return(NULL_CONV_ERR);
  }
  /*
   * Check if convertion is possible ...
   */
  if((status=CheckConv(srcdat->kind,srcdat->flag,srcdat->kind,flag)) != OK) {
    gerror(status,"ConvertData(): convertion is impossible");
    return(status);
  }
  if((status=CreateSPData(tgtdat,srcdat->kind,srcdat->dim,srcdat->n,flag)) != OK) {
    gerror(status,"ConvertData(): cannot create output data structure");
    return(status);
  }
  /*
   * Run convertion ...
   */
  if(flag & WITHZ)
    DataZeroMean(srcdat);
  
  srcl=DataVecSize(srcdat);
  tgtl=DataVecSize(*tgtdat);

  for(i=0;i < srcdat->n;i++) { /* foreach frame */
    srcs=DataGetVec(srcdat,i,srcl);
    tgts=DataGetVec(*tgtdat,i,tgtl);

    /***** Copy static coefficients *****/
    for(j=0,k=0;j < srcdat->dim;j++,k++)
      tgts[k]=srcs[k];
    if((flag & WITHE) && !(flag & WITHN)) {
      tgts[k]=srcs[k];
      k++;
    }

    /***** Add delta *****/
    if(flag & WITHD) {
      /* Get the right vectors (duplicating boundary vectors) */
      p2=DataGetVec(srcdat,i-2 > 0?i-2:0,srcl);
      p1=DataGetVec(srcdat,i-1 > 0?i-1:0,srcl);
      n1=DataGetVec(srcdat,i+1 < srcdat->n?i+1:srcdat->n-1,srcl);
      n2=DataGetVec(srcdat,i+2 < srcdat->n?i+2:srcdat->n-1,srcl);
      /* Apply delta to static coeffs */
      for(j=0;j < srcdat->dim;j++,k++)
	tgts[k]=(n1[j]-p1[j]+2.0*(n2[j]-p2[j]))/10.0;
      /* Apply delta to energy (if energy) 
         Note that j=srcdat->dim and corresponds to the energy */
      if(flag & WITHE) {
	tgts[k]=(n1[j]-p1[j]+2.0*(n2[j]-p2[j]))/10.0;
	k++;
      }
    }
  }

  /***** Add delta-delta *****/
  if(flag & WITHA) {
    for(i=0;i < srcdat->n;i++) {
      tgts=DataGetVec(*tgtdat,i,tgtl);
      p=((flag & WITHE) && !(flag & WITHN)) ? srcdat->dim+1:srcdat->dim;
      p2=DataGetVec(*tgtdat,i-2 > 0?i-2:0,tgtl);
      p1=DataGetVec(*tgtdat,i-1 > 0?i-1:0,tgtl);
      n1=DataGetVec(*tgtdat,i+1 < (*tgtdat)->n?i+1:(*tgtdat)->n-1,tgtl);
      n2=DataGetVec(*tgtdat,i+2 < (*tgtdat)->n?i+2:(*tgtdat)->n-1,tgtl);
      /* Apply delta to delta coeffs */
      for(j=0,k=(flag & WITHE)?p+srcdat->dim+1:p+srcdat->dim;j < srcdat->dim;j++,k++)
	tgts[k]=(n1[p+j]-p1[p+j]+2.0*(n2[p+j]-p2[p+j]))/10.0;
      /* Apply delta to delta energy (if energy) 
	 Note that p+j now points to deltaE in tgts[] */
      if(flag & WITHE)
	tgts[k]=(n1[p+j]-p1[p+j]+2.0*(n2[p+j]-p2[p+j]))/10.0;
    }
  }

  DONE;
}

/* ---------------------------------------------------------------- */
/* ----- err_t CheckConv(datakind_t,flag_t,datakind_t,flag_t) ----- */
/* ---------------------------------------------------------------- */
/*
 * Check that convertion from (skind,sflag) to (tkind,tflag) is possible.
 * I'm sure I forgot some cases that may cause some problems. Just add them
 * whenever you find one. Thank you!
 *
 */
err_t CheckConv(datakind_t skind,flag_t sflag,datakind_t tkind,flag_t tflag)
{
  if(tkind != skind) {
    gerror(CONV_ERR,"CheckConv(): cannot convert from %s to %s",
	   kind2str(skind),kind2str(tkind));
    return(CONV_ERR);
  }
  if((tflag & WITHE) && !(sflag & WITHE)) {
    gerror(CONV_ERR,"CheckConv(): target has energy but source has no energy");
    return(CONV_ERR);
  }
  if((tflag & WITHN) && !(sflag & WITHE)) {
    gerror(CONV_ERR,"CheckConv(): target has log E suppressed but source has no energy");
    return(CONV_ERR);
  }
  if((tflag & WITHN) && !(tflag & WITHE)) {
    gerror(CONV_ERR,"CheckConv(): target has log E suppressed but no energy");
    return(CONV_ERR);
  }
  if((tflag & WITHN) && !(tflag & WITHD)) {
    gerror(CONV_ERR,"CheckConv(): target has log E suppressed but no delta");
    return(CONV_ERR);
  }
  if((tflag & WITHA) && !(tflag & WITHD)) {
    gerror(CONV_ERR,"CheckConv(): cannot add acceleration without delta");
    return(CONV_ERR);
  }
  DONE;
}

/* --------------------------------------- */
/* ----- flag_t ConvStr2Flag(char *) ----- */
/* --------------------------------------- */
flag_t ConvStr2Flag(char *str)
{
  char *ptr=str;
  flag_t flag=(flag_t)0;
  
  while(*ptr) {
    switch(toupper(*ptr)) {
    case 'D': flag |= WITHD; break;
    case 'A': flag |= WITHA; break;
    case 'Z': flag |= WITHZ; break;
    case 'N': flag |= WITHN; break;
    default:
      gerror(CONV_FLAG_WARN,"ConvStr2Flag(): unknown convertion character %c (ignoring)",
	     toupper(*ptr));
    }
    ptr++;
  }
  return(flag);
}

#undef _spro_conv_c_
