/************************************************************/
/*                        spro_fft.c                        */
/*                                                          */
/*                       SPro Library                       */
/* -------------------------------------------------------- */
/* Guig                                           Apr. 1997 */
/************************************************************/
/*
 * CVS log:
 *
 * $Author: gravier $
 * $Date: 1999/12/03 13:23:44 $
 * $Revision: 1.3 $
 *
 */
/*
 * Procedures related to filter bank analysis of speech signal
 * and any other kind of FFT.
 *
 * To be fair, I must admit that this part of the library
 * mainly relies on the Numerical Recipes implementation of the
 * FFT (four1()) and real FFT (realfft()).
 *
 * Please report to spro.h header for brief description of the
 * functions implemented here. Implementation issues are given
 * in comments before each function source code.
 *
 * Have fun and enjoy!
 */

#define _spro_fft_c_

#include <spro.h>

#define OM_EPSILON 0.00001
#define PI2 6.28318530717959 

/* -------------------------------------------------------------------------- */
/* ----- err_t FFTransform(sig_t *,unsigned long,vector_t *,vector_t *) ----- */
/* -------------------------------------------------------------------------- */
/*
 * Performs FFT on "signal" x:
 *
 *    - Truncate or pad with zeros if npts != in->n
 *    - Copy truncated/padded x->n to real part and set imaginary part to 0.0
 *    - Call mixfft [fft()]
 *    - Write phase and modulus to the output vectors
 *
 * Either modul or phase can be NULL if one is not
 * interested in this result.
 *
 * Note that "signal" may not be a signal, but it can be 
 * LPCOEFF vector hidden in a sig_t structure!
 */
err_t FFTransform(sig_t *x,unsigned long npts,vector_t *modul,vector_t *phase)
{
  int i,j;
  float tmp[MAX_FFT_SIZE];

  /*
   * Check that (nearly) all is OK!
   */
  if(npts > MAX_FFT_SIZE) {
    gerror(FFT_LENGTH_ERR,"FFTransform(): FFT length too large (npts= %d, max= %d)",
	   npts,MAX_FFT_SIZE);
    return(FFT_LENGTH_ERR);
  }
  if(frexp((double)npts,&i) != (double)0.5) {
    gerror(FFT_LENGTH_ERR,"FFTransform(): FFT length (%d) is not a power of two",
	   npts);
    return(FFT_LENGTH_ERR);
  }
  if(x->n == 0 || npts == 0) {
    gerror(ZERO_LENGTH_WARN,"FFTransform(): zero length signal (nb. samples= %d, npts= %d)",
	   x->n,npts);
    DONE;
  }
  if(modul)
    if(modul->n < npts/2) {
      gerror(OUT_SIZE_ERR,"FFTransform(): not enough memory allocated to module vector");
      return(OUT_SIZE_ERR);
    }
  if(phase)
    if(phase->n < npts/2) {
      gerror(OUT_SIZE_ERR,"FFTransform(): not enough memory allocated to phase vector");
      return(OUT_SIZE_ERR);
    }
  if(x->n > npts)
    gerror(FFT_TRUNC_WARN,"FFTransform(): signal length greater than FFT order (truncating)");

  /*
   * Copy signal to buffers and apply fft
   */
  for(i=0;i<x->n && i<npts;i++)
    tmp[i]=(float)x->s[i];
  for(;i<npts;i++)
    tmp[i]=(float)0.0;

  realfft(tmp,npts);

  /*
   * Fill modul and/or phase vetor(s)
   */
  for(i=0,j=0;i < npts;i+=2,j++) {
    if(modul)
      modul->data[j]=(param_t)sqrt(tmp[i]*tmp[i]+tmp[i+1]*tmp[i+1]);
    if(phase)
      phase->data[j]=(param_t)atan(tmp[i+1]/tmp[i]);
  }
  if(modul)
    modul->n=npts/2;
  if(phase)
    phase->n=npts/2;
  DONE;
}

/* ------------------------------------------------------------------------ */
/* -----err_t LogFilterBank(sig_t *,float,int,float,float,vector_t *) ----- */
/* ------------------------------------------------------------------------ */
/*
 * Apply a triangular filter-bank in the range [fmin,fmax]. The filter
 * cut-off frequencies are regularly spaced on the \alpha-transformed
 * domain. Note that fmin and fmax are given as normalized frequencies
 * (i.e. f \in [0,0.5]). If fmin == fmax, the full band is considered.
 *
 * Cut-off frequencies are updated when either alpha or nfilt.
 */
err_t LogFilterBank(sig_t *x,float alpha,int nfilt,float f_min,float f_max,vector_t *v)
{
  int i;
  param_t tmp[FBANK_FFT_SIZE/2];           /* Data for module vector             */
  vector_t module;                         /* Module of FFT'ed signal            */
  float o;                                 /* Current cut-off frequency          */
  static int p_index[MAX_BANK_FILTER+2];   /* Indexes of cut-off pulses          */
  float deltao;                            /* delta omega                        */
  static float alpha_mem=10.0;             /* What was alpha at last call?       */
  static float nfilt_mem=0;                /* What was nfilt at last call?       */
  float omin,omax;
  err_t status;

  /*
   * Check parameters 
   */
  if(f_min < 0.0 || f_max < 0.0 || f_min > 0.5 || f_max > 0.5 || f_max < f_min) {
    gerror(FREQ_BOUND_ERR,"LogFilterBank(): invalid min and/or max frequencies [%.2f,%.2f])",f_min,f_max);
    return(FREQ_BOUND_ERR);
  }
  if(f_max == f_min) {
    f_min=0.0;
    f_max=0.5;
  }
  if(nfilt < MIN_BANK_FILTER || nfilt > MAX_BANK_FILTER) {
    gerror(FBANK_ORDER_ERR,"LogFilterBank(): invalid number of filters n=%d (min= %d,max= %d)",
	   nfilt,MIN_BANK_FILTER,MAX_BANK_FILTER);
    return(FBANK_ORDER_ERR);
  }
  if(v->n < nfilt) {
    gerror(OUT_SIZE_ERR,"LogFilterBank(): not enough memory allocated for output vector");
    return(OUT_SIZE_ERR);
  }
  /*
   * Run FFT
   */
  module.data=tmp;
  module.n=FBANK_FFT_SIZE/2;
  if((status=FFTransform(x,FBANK_FFT_SIZE,&module,NULL)) != OK) {
    gerror(status,"LogFilterBank(): FFT error");
    FreeVector(&module);
    return(status);
  }
  /*
   * Compute cut-off frequencies.
   */
  /* compute initial and final pulses (in the transform domain) */
  if(alpha != 0.0) {
    omin=Theta(2.0*M_PI*f_min,alpha);
    omax=Theta(2.0*M_PI*f_max,alpha);
  }
  else {
    omin=2.0*M_PI*f_min;
    omax=2.0*M_PI*f_max;
  }
  /* set initial and final indexes */
  p_index[0]=round(omin*(float)(module.n-1)/M_PI);
  p_index[nfilt+1]=round(omax*(float)(module.n-1)/M_PI);
  /* (re)set central cut-off values */
  if(nfilt_mem != nfilt || alpha_mem != alpha) {
    nfilt_mem=nfilt;
    alpha_mem=alpha;
    deltao=(omax-omin)/(float)(nfilt+1);
    o=omin;
    for(i=0;i < nfilt;i++) {
      o += deltao;
      if(alpha != 0.0)
	o=ThetaInv(o,alpha);
      p_index[i+1]=round(o*(float)(module.n-1)/M_PI);
    }
  }
  /*
   * Apply log-filter bank
   */
  LogTriangularFBank(&module,nfilt,p_index,v);

  DONE;
}

/* ---------------------------------------------------------------- */
/* ----- err_t LogMelFilterBank(sig_t *,float,int,vector_t *) ----- */
/* ---------------------------------------------------------------- */
/*
 * Apply a triangular filter-bank in the range [fmin,fmax]. The filter
 * cut-off frequencies are regularly spaced on the Mel frequency scale.
 * If f_min == f_max, the full band is considered.
 *
 * Cut-off frequencies are updated when either nfilt or freq.
 */
err_t LogMelFilterBank(sig_t *x,float freq,int nfilt,float f_min,float f_max,vector_t *v)
{
  int i;
  param_t tmp[FBANK_FFT_SIZE/2];           /* Data for module vector             */
  vector_t module;                         /* Module of FFT'ed signal            */
  static int p_index[MAX_BANK_FILTER+2];   /* Indexes of cut-off frequencies     */
  float deltaf;                            /* delta f between two filters        */
  static float freq_mem=0.0;               /* What was freq at last call?        */
  static int nfilt_mem=0;                  /* What was nfilt at last call?       */
  err_t status;

  /*
   * Check parameters 
   */
  if(f_min < 0.0 || f_max < 0.0 || f_min > freq/2.0 || f_max > freq/2.0 || f_max < f_min) {
    gerror(FREQ_BOUND_ERR,"LogFilterBank(): invalid min and/or max frequencies [%.2f,%.2f])",f_min,f_max);
    return(FREQ_BOUND_ERR);
  }
  if(f_max == f_min) {
    f_min=0.0;
    f_max=freq/2.0;
  }
  if(nfilt < MIN_BANK_FILTER || nfilt > MAX_BANK_FILTER) {
    gerror(FBANK_ORDER_ERR,"LogMelFilterBank(): invalid number of filters n=%d (min= %d,max= %d)",
	   nfilt,MIN_BANK_FILTER,MAX_BANK_FILTER);
    return(FBANK_ORDER_ERR);
  }
  if(v->n < nfilt) {
    gerror(OUT_SIZE_ERR,"LogMelFilterBank(): not enough memory allocated for output vector");
    return(OUT_SIZE_ERR);
  }
  /*
   * Run FFT
   */  
  module.data=tmp;
  module.n=FBANK_FFT_SIZE/2;
  if((status=FFTransform(x,FBANK_FFT_SIZE,&module,NULL)) != OK) {
    gerror(status,"LogMelFilterBank(): FFT error");
    FreeVector(&module);
    return(status);
  }
  /*
   * Compute cut-off frequencies.
   */
  /* compute initial and final pulses (in the transform domain) */
  p_index[0]=round(2.0*(module.n-1)*f_min/freq);
  p_index[nfilt+1]=round(2.0*(module.n-1)*f_max/freq);
  if(nfilt_mem != nfilt || freq_mem != freq) {
    nfilt_mem=nfilt;
    freq_mem=freq;
    deltaf=Mel(freq/2.0)/(float)(nfilt+1);
    for(i=0;i<nfilt;i++)
      p_index[i+1]=round(2.0*(module.n-1)*MelInv(f_min+(i+1)*deltaf)/freq);
  }
  /*
   * Apply log-filter bank
   */
  LogTriangularFBank(&module,nfilt,p_index,v);

  DONE;
}

/* -------------------------------------------------------- */
/* ----- err_t DCTransform(vector_t *,int,vector_t *) ----- */
/* -------------------------------------------------------- */
/*
 * Apply a p order DCT to vector v1.
 * Results are returned through v2.
 *
 * If {m[1],...,m[N]} are the output of the filters, then
 *    c[i]=sqrt(2/N)*sum for j=1 to N of(m[j]cos(M_PI*i*(j-0.5)/N) i=1,...,p
 *
 * This is what is implemented here with arrays indexed from 0 to N-1.
 *
 */
err_t DCTransform(vector_t *v1,int p,vector_t *v2)
{
  int i,j;
  static unsigned long vlength=0;
  static int dct_order=0;
  static float dct_kernel[MAX_STATIC_SIZE][MAX_BANK_FILTER];
  static float dct_norm;

  if(p < 1 || p > MAX_STATIC_SIZE) {
    gerror(DCT_ORDER_ERR,"DCTransform(): DCT order (%d)is out of range (min= 1, max= %d)",
	   p,MAX_STATIC_SIZE);
    return(DCT_ORDER_ERR);
  }
  if(v1->n < p || v1->n > MAX_BANK_FILTER) {
    gerror(DCT_LENGTH_ERR,"DCTransform(): cannot apply a %d-order DCT to %d dimension vector",
	   p,v1->n);
    return(DCT_LENGTH_ERR);
  }
  if(v2->n < p) {
    gerror(OUT_SIZE_ERR,"DCTransform(): not enough memory in output vector");
    return(OUT_SIZE_ERR);
  }

  if(vlength != v1->n || dct_order != p)
    dct_norm=(float)sqrt(2.0/(float)(v1->n));
  for(i=1;i<=p;i++) {
    v2->data[i-1]=0.0;
    for(j=1;j<=v1->n;j++) {
      if(vlength != v1->n || dct_order != p)
	dct_kernel[i-1][j-1]=(float)cos(M_PI*i*(j-0.5)/(float)(v1->n));
      v2->data[i-1] += (param_t)(v1->data[j-1]*dct_kernel[i-1][j-1]);
    }
    v2->data[i-1] *= (param_t)dct_norm;
  }
  v2->n=p;
  vlength=v1->n;
  dct_order=p;
  DONE;
}

        /******************************/
        /***** PRIVATE PROCEDURES *****/
        /******************************/

/* -------------------------------------------------------------------- */
/* ----- void LogTriangularFBank(vector_t *,int,int *,vector_t *) ----- */
/* -------------------------------------------------------------------- */
/*
 * Apply triangular filter bank to module vector and return the log of
 * the energy in each band. Table p_index contains the indexes of the
 * cut-off frequencies. Looks like something like this:
 * 
 *                      filter 2   
 *                   <------------->
 *                filter 1           filter 3
 *             <----------->       <------------->
 *        | | | | | | | | | | | | | | | | | | | | | ..........
 *         0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9  ..........
 *             ^     ^     ^       ^             ^
 *             |     |     |       |             |
 *          p_in[0]  |  p_in[2]    |          p_in[4]
 *                p_in[1]       p_in[3]  
 *
 */
void LogTriangularFBank(vector_t *module,int nfilt,int *p_index,vector_t *v)
{
  int i,j;
  float a;
  double res;

  for(i=0;i < nfilt;i++) {
    /*
     *     from j=p_in[i] to p_in[i+1]-1  ascending  step: a=1/(p_in[i+1]-p_in[i]+1)
     *     from j=p_in[i+1] to p_in[i+2]  descending step: a=1/(p_in[i+2]-p_in[i+1]+1)
     */
    res=0.0;
    a=1.0/(float)(p_index[i+1]-p_index[i]+1);
    for(j=p_index[i];j < p_index[i+1];j++)
      res += module->data[j]*(1.0-a*(p_index[i+1]-j));
    a=1.0/(float)(p_index[i+2]-p_index[i+1]+1);
    for(j=p_index[i+1];j<=p_index[i+2];j++)
      res += module->data[j]*(1.0-a*(j-p_index[i+1]));
    v->data[i]=(res < FBANK_OUT_FLOOR)?(param_t)log(FBANK_OUT_FLOOR):(param_t)log(res);
  }
  v->n=nfilt;
}

/* ----------------------------------------------------------------- */
/* ----- void TriangularFBank(vector_t *,int,int *,vector_t *) ----- */
/* ----------------------------------------------------------------- */
/*
 * Same as LogTriangularFBank() except that it does not return the log
 */
void TriangularFBank(vector_t *module,int nfilt,int *p_index,vector_t *v)
{
  int i,j;
  float a;
  double res;

  for(i=0;i < nfilt;i++) {
    /*
     *     from j=p_in[i] to p_in[i+1]-1  ascending  step: a=1/(p_in[i+1]-p_in[i]+1)
     *     from j=p_in[i+1] to p_in[i+2]  descending step: a=1/(p_in[i+2]-p_in[i+1]+1)
     */
    res=0.0;
    a=1.0/(float)(p_index[i+1]-p_index[i]+1);
    for(j=p_index[i];j < p_index[i+1];j++)
      res += module->data[j]*(1.0-a*(p_index[i+1]-j));
    a=1.0/(float)(p_index[i+2]-p_index[i+1]+1);
    for(j=p_index[i+1];j<=p_index[i+2];j++)
      res += module->data[j]*(1.0-a*(j-p_index[i+1]));
    v->data[i]=(param_t)res;
  }
  v->n=nfilt;
}

/* ------------------------------------ */
/* ----- float Theta(float,float) ----- */
/* ------------------------------------ */
/*
 * The function determines the pulsation value on the transformed
 * axis (omega), assuming a spectral factor alpha.
 * 
 * For details on this, refer to:
 *       "Analyse spectrale a resolution variable", Ph.D. Thesis,
 *       Christian Chouzenoux, ENST
 * In this Ph.D., we are given omega'=Theta(omega).
 */
float Theta(float o,float a)
{
  double v;

  v=atan(((1.0-a*a)*sin(o))/((1.0+a*a)*cos(o)-2.0*a));
  return((float)(v < (double)0.0?v+M_PI:v));
}

/* --------------------------------------- */
/* ----- float ThetaInv(float,float) ----- */
/* --------------------------------------- */
/*
 * The function determines the pulsation value on the original
 * axis (omega), assuming a spectral factor alpha.
 * 
 * For details on this, refer to:
 *       "Analyse spectrale a resolution variable", Ph.D. Thesis,
 *       Christian Chouzenoux, ENST
 *
 * In this Ph.D., we are given omega'=Theta(omega). As I don't feel
 * like finding out an analytical solution to the inverse problem,
 * I simply look for a dichotomic solution to the problem by using
 * omega in the range [0,PI], as the transformation is monotonous.
 *
 * oop: original omega'
 * op:  omega'
 * opmem: omega' memory
 *
 * NOTE: If someone feels like programming an analytical solution
 * to the problem, he is welcome! Have fun ...
 *
 */
float ThetaInv(float oop,float alpha)
{
  float o,op,oinf,osup;
  double val;

  if(oop <= 0.0)
    return(0.0);
  if(oop >= M_PI)
    return(M_PI);
  oinf=0.0;
  osup=M_PI;
  do {
    o=oinf+(osup-oinf)/2.0;
    val=atan(((1.0-alpha*alpha)*sin(o))/((1.0+alpha*alpha)*cos(o)-2.0*alpha));
    op=val < 0?(float)val+M_PI:(float)val;
    if(op > oop)
      osup=o;
    else
      oinf=o;
  }
  while((float)fabs(oop-op) > OM_EPSILON);
  return(o);
}

/* ---------------------------- */
/* ----- float Mel(float) ----- */
/* ---------------------------- */
float Mel(float f)
{
  return(2595.0*log10(1+f/700.0));
}

/* ------------------------------- */
/* ----- float MelInv(float) ----- */
/* ------------------------------- */
/*
 * Return the value corresponding to Mel frequency f on a 
 * linear scale.
 *   Mel(f)=2595*log10(1+f/700)
 *   MelInv(f)=700*(10^(f/2595)-1)
 */
float MelInv(float f)
{
  return((float)(700.0*(pow(10,f/2595.0)-1)));
}

/*
 * FFT and real FFT copied from the NR in C.
 *
 * See http://cfatab.harvard.edu/nr/bookc.html, chapter 12.
 *
 */
/* ------------------------------------------- */
/* ----- void fft(float *,unsigned long) ----- */
/* ------------------------------------------- */
/*
 * Adapted from NR to match arrays indexed from 0 and not from 1
 */
void fft(float *x,unsigned long nn)
{
  unsigned long n,mmax,m,j,istep,i;
  double wtemp,wr,wpr,wpi,wi,theta;
  float tempr,tempi;

  n=nn << 1;
  j=1;
  for(i=1;i<n;i+=2) {
    if(j > i) {
      tempr=x[j-1];
      x[j-1]=x[i-1];
      x[i-1]=tempr;
      tempi=x[j];
      x[j]=x[i];
      x[i]=tempi;
    }
    m=n >> 1;
    while(m >= 2 && j > m) {
      j -= m;
      m >>= 1;
    }
    j += m;
  }
  mmax=2;
  while(n > mmax) {
    istep=mmax << 1;
    theta=PI2/mmax;
    wtemp=sin(0.5*theta);
    wpr=-2.0*wtemp*wtemp;
    wpi=sin(theta);
    wr=1.0;
    wi=0.0;
    for(m=1;m<mmax;m+=2) {
      for(i=m;i<=n;i+=istep) {
	j=i+mmax;
	tempr=wr*x[j-1]-wi*x[j];
	tempi=wr*x[j]+wi*x[j-1];
	x[j-1]=x[i-1]-tempr;
	x[j]=x[i]-tempi;
	x[i-1] += tempr;
	x[i] += tempi;
      }
      wtemp=wr;
      wr += (wtemp*wpr-wi*wpi);
      wi += (wi*wpr+wtemp*wpi);
    }
    mmax=istep;
  }
}

/* ----------------------------------------------- */
/* ----- void realfft(float *,unsigned long) ----- */
/* ----------------------------------------------- */
/*
 * Adapted from NR to match arrays indexed from 0 and not from 1
 */
void realfft(float *x,unsigned long n)
{
  unsigned long i,i1,i2,i3,i4,np3;
  float c1=0.5,c2=-0.5,h1r,h2r,h1i,h2i;
  double wpr,wpi,wr,wi,wtemp,theta;

  theta=M_PI/(double)(n >> 1);
  fft(x,n >> 1);
  wtemp=sin(0.5*theta);
  wpr=-2.0*wtemp*wtemp;
  wpi=sin(theta);
  wr=1.0+wpr;
  wi=wpi;
  np3=n+3;
  for(i=2;i<=(n>>2);i++) {
    /* i4=1+(i3=np3-(i2=1+(i1=i+i-1))); */
    i1=i+i-1;
    i2=1+i1;
    i3=np3-i2;
    i4=1+i3;
    h1r=c1*(x[i1-1]+x[i3-1]);
    h1i=c1*(x[i2-1]-x[i4-1]);
    h2r=-c2*(x[i2-1]+x[i4-1]);
    h2i=c2*(x[i1-1]-x[i3-1]);
    x[i1-1]=h1r+wr*h2r-wi*h2i;
    x[i2-1]=h1i+wr*h2i+wi*h2r;
    x[i3-1]=h1r-wr*h2r+wi*h2i;
    x[i4-1]=-h1i+wr*h2i+wi*h2r;
    wtemp=wr;
    wr += (wtemp*wpr-wi*wpi);
    wi += (wi*wpr+wtemp*wpi);
  }
  h1r=x[0];
  x[0] += x[1];
  x[1]=h1r-x[1];
}

#undef _spro_fft_c_
