/************************************************************/
/*                     spro_global.c                        */
/*                                                          */
/*                      SPro Library                        */
/* -------------------------------------------------------- */
/* Guig                                           Apr. 1997 */
/************************************************************/
/*
 * CVS log:
 *
 * $Author: gravier $
 * $Date: 1999/02/27 15:47:32 $
 * $Revision: 1.4 $
 *
 */
/*
 * General purpose signal processing tools and data basic
 * operations. 
 * Most of those functions were defined in spro.c (SPro v2.0).
 * Nothing radically changed!
 * 
 */

#define _spro_global_c_

#include <spro.h>

          /********** MISCELLANEOUS PROCEDURES **********/

/* -------------------------------------- */
/* ----- char *kind2str(datakind_t) ----- */
/* -------------------------------------- */
char *kind2str(datakind_t kind)
{
  static char *kindname[]={
    "OTHER",
    "FBANK",
    "FBCEPSTRA",
    "LPCEPSTRA",
    "LPCOEFF",
    "PARCOR",
    "LAR"
  };

  return(kindname[kind]);
}

/* -------------------------------- */
/* ----- char *win2str(win_t) ----- */
/* -------------------------------- */
char *win2str(win_t win)
{
  static char *windname[]={
    "None",
    "Hamming",
    "Hanning",
    "Blackman"
  };

  return(windname[win]);
}

          /*********** DATA RELATED PROCEDURE **********/

/* ----------------------------------------------- */
/* ----- unsigned long DataVecSize(data_t *) ----- */
/* ----------------------------------------------- */
/*
 * Return the number of elements in a feature vector
 */
unsigned long DataVecSize(data_t *data)
{
  unsigned long n;

  n=data->dim;
  if(data->flag & WITHE)
    n++;
  if(data->flag & WITHD)
    if(data->flag & WITHA)
      n *= 3;
    else
      n += n;
  if(data->flag & WITHN) 
    n--;
  return(n);
}

/* ----------------------------------------------------------- */
/* ----- param_t *DataGetVec(data_t *,int,unsigned long) ----- */
/* ----------------------------------------------------------- */
/*
 * Returns a pointer to the n'th d length feature vector in data.
 * Vectors start at index 0.
 */
param_t *DataGetVec(data_t *data,int n,unsigned long nelem)
{
  if(n < 0 || n >= data->n) {
    gerror(VECTOR_ID_ERR,"DataGetVec(): invalid vector index %d (max= %d)",n,data->n-1);
    return(NULL);
  }
  return(&(data->s[n*nelem]));
}

/* --------------------------------------- */
/* ----- void DataZeroMean(data_t *) ----- */
/* --------------------------------------- */
/*
 * Set mean of static coefficients to 0.
 * If delta and/or acc already exist, they are not
 * recomputed.
 */
void DataZeroMean(data_t *data)
{
  int i,j;
  length_t vecsize=DataVecSize(data);
  param_t m[MAX_STATIC_SIZE];
  param_t *x;

  if(data->flag & WITHZ)
    return;
  
  for(j=0;j < data->dim;j++)
    m[j]=0.0;
  for(i=0;i < data->n;i++) {
    x=DataGetVec(data,i,vecsize);
    for(j=0;j < data->dim;j++)
      m[j] += x[j];
  }
  for(j=0;j < data->dim;j++)
    m[j] /= (float)(data->n);
  for(i=0;i < data->n;i++) {
    x=DataGetVec(data,i,vecsize);
    for(j=0;j<data->dim;j++)
      x[j] -= m[j];
  }

  data->flag |= WITHZ;
  return;
}

          /********** CEPSTRAL ANALYSIS PROCEDURES **********/

/* ------------------------------------------ */
/* ----- void CepLifter(vector_t *,int) ----- */
/* ------------------------------------------ */
/*
 * Apply liftering according to:
 *
 *  v->data[i-1]=(1.0+l*sin(i*M_PI/l)/2.0)*v->data[i-1]
 *  for i=1 to v->n.
 */
void CepLifter(vector_t *v,int l)
{
  int i;
  static float weight[MAX_CEPSTRUM];
  static int lmem=0.0;

 if(l == 0)
   return;
 if(l != lmem) {
   lmem=l;
   for(i=1;i<=MAX_CEPSTRUM;i++)
     weight[i-1]=1.0+(float)l*sin(i*M_PI/(float)l)/2.0;
 }
 for(i=0;i<v->n && i<MAX_CEPSTRUM;i++)
   v->data[i] *= weight[i];
}

/* -------------------------------------------- */
/* ----- void ScaleEnergy(data_t *,float) ----- */
/* -------------------------------------------- */
void ScaleEnergy(data_t *p,float s)
{
  unsigned long vecsize=DataVecSize(p);
  param_t max,*e;
  int i;
  
  e=p->s+vecsize-1;
  max=*e;
  e += vecsize;
  for(i=1;i < p->n;i++,e += vecsize)
    if(*e > max)
      max=*e;
  e=p->s+vecsize-1;
  for(i=0;i < p->n;i++,e += vecsize)
    *e=(*e - max)*s+1.0;
}

          /********** SIGNAL PROCESSING PROCEDURES **********/

/* ---------------------------------------------------------- */
/* ----- err_t SigGetFrame(sig_t *,int,int,int,sig_t *) ----- */
/* ---------------------------------------------------------- */
/*
 * Return signal corresponding to nth frame in sig.
 * Frames are l samples long, every d samples.
 * First frame is the number 1.
 * The frame signal is returned through frame variable, which must
 * be allocated before, and the return value is true as long as
 * a frame has been read.
 */
err_t SigGetFrame(sig_t *sig,int nth,int l,int d,sig_t *frame)
{
  int i,from,to;

  if(frame->n < l) {
    gerror(OUT_SIZE_ERR,"SigGetFrame(): not enough memory in output frame structure");
    return(OUT_SIZE_ERR);
  }
  from=(nth-1)*d;
  to=from+l;
  if(to > sig->n || nth < 1)
    return(FRAME_ID_ERR);
  for(i=from;i<to;i++)
    frame->s[i-from]=sig->s[i];
  frame->n=l;
  DONE;
}

/* ------------------------------------------- */
/* ----- float SigCenter(sig_t *,bool_t) ----- */
/* ------------------------------------------- */
/*
 * Compute mean for sig and remove it if flag is set to true.
 * Return mean value of signal in any case.
 */
float SigCenter(sig_t *sig,bool_t doit)
{
  int i;
  double mean=0.0;

  for(i=0;i<sig->n;i++)
    mean+=(double)(sig->s[i]);
  mean=mean/(float)(sig->n);
  if(doit)
    for(i=0;i<sig->n;i++)
      sig->s[i] -= (sample_t)mean;
  return((float)mean);
}

/* ----------------------------------------- */
/* ----- float SigNorm(sig_t *,bool_t) ----- */
/* ----------------------------------------- */
/*
 * Compute sqrt(sum(sig->s[i]^2)).
 * If flag is set to true, every sample is divided
 * by this value so that it belongs to [-1,1].
 * Return the gain in any case.
 */
float SigNorm(sig_t *sig,bool_t flag)
{
  int i;
  double gain=0.0;

  for(i=0;i<sig->n;i++)
    gain+=sig->s[i]*sig->s[i];
  gain=sqrt(gain);
  if(flag && gain != 0.0)
    for(i=0;i<sig->n;i++)
      sig->s[i] /= (sample_t)gain;
  return((float)gain);
}

/* ------------------------------------------ */
/* ----- err_t SigWeight(sig_t *,win_t) ----- */
/* ------------------------------------------ */
/*
 * Apply a weighting window to the signal.
 * No error is returned but only a warning is displayed
 * if the window is not a valid one.
 *
 * Pre-computed windows can speed up the stuff. Let's do it!
 */
err_t SigWeight(sig_t *sig,win_t window)
{
  int i;
  bool_t setwin=false;
  static sample_t win[MAX_SAMPLESINFRAME];
  static int winlgth=0;
  static win_t winkind;

  /*
   * Check if the weighting coefficients are already computed
   */
  switch(window) {
  case RECTANGLE:
    break;
  case HAMMING:
    if(sig->n != winlgth || winkind != HAMMING) {
      winlgth=sig->n;
      winkind=HAMMING;
      setwin=true;
    }
    for(i=0;i<sig->n;i++) {
      if(setwin)
	win[i]=(sample_t)(0.54-0.46*cos(2*M_PI*i/(winlgth-1)));
	sig->s[i]=sig->s[i]*win[i];
    }
    break;
  case HANNING:
    if(sig->n != winlgth || winkind != HANNING) {
      winlgth=sig->n;
      winkind=HANNING;
      setwin=true;
    }
    for(i=0;i<sig->n;i++) {
      if(setwin)
	win[i]=(sample_t)(0.5*(1-cos(2*M_PI*i/(winlgth-1))));
      sig->s[i]=sig->s[i]*win[i];
    }
    break;
  case BLACKMAN:
    gerror(NYI_ERR,"SigWeight(): blackman windows not yet supported");
    return(NYI_ERR);
    break;
  default:
    gerror(WIN_ID_ERR,"SigWeight(): invalid weighting window");
    return(WIN_ID_ERR);
  }
  DONE;
}

/* ---------------------------------------- */
/* ----- err_t SigEmph(sig_t *,float) ----- */
/* ---------------------------------------- */
/*
 *               ------------
 *               |          |
 *     x->s ---->| 1-a.z^-1 |----> y->s
 *               |          |
 *               ------------
 *
 * If a>0: y[t]=x[t]-a.x[t-1], with x[t]=0 for t<0
 * else:   y[t]=x[t]+|a|.y[t-1], with y[t]=0 for t<0
 *
 * If a is not in [-1,1], a warning is printed.
 *
 */
err_t SigEmph(sig_t *sig,float a)
{
  sample_t xprev,xcur;
  int i;

  if(a <= -1.0 || a >= 1.0) {
    gerror(EMPHCO_ERR,"SigEmph(): invalid emphasis coefficient %.2f",a);
    return(EMPHCO_ERR);
  }
  if(a != 0.0) {
    xprev=sig->s[0];
    for(i=1;i<sig->n;i++) {
      if(a>0) {
	xcur=sig->s[i];
	sig->s[i]=xcur-a*xprev;
	xprev=xcur;
      }
      else
	sig->s[i]=sig->s[i]-a*sig->s[i-1];  /* |a| = -a since a < 0 */
    }
  }
  DONE;
}

/* ----------------------------------------------------- */
/* ----- err_t SigDownSample(sig_t **,sig_t *,int) ----- */
/* ----------------------------------------------------- */
/*****
 * NOTE: Decimation filter is DFILTER_S coefficients wide
 *       See dfilter.h for more details
 *
 * Down sample isig "howmany" times and writes result to osig.
 * Note that it is the only function where osig is allocated.
 * This is not very clean but it avoids precomputing the number
 * of samples in the down sampled signal!
 *
 * The function operates as follow:
 *      Copy original signal in buf2.
 *      buf1 recieves the low pass filtered signal which is decimated
 *      in buf2, ans so on ...
 *      Finally, buf2 is copied to out->s
 *
*****/

#include <dfilter.h>

err_t SigDownSample(sig_t **out,sig_t *sig,int howmany)
{
  sample_t *buf1;     /* buffer for filtered signal               */
  sample_t *buf2;     /* buffer for decimated signal              */
  int i,j;            /* Always those loop variables              */
  int lgth;           /* Number of samples in buffers             */
  int ndecim=0;       /* Number of decimation performed           */
  int from;
  err_t status;

  if((buf1=(sample_t *)malloc(sig->n*sizeof(sample_t))) == NULL) {
    gerror(ALLOC_ERR,"SigDownSample(): cannot allocate %d bytes for buffer #1",
	   sig->n*sizeof(sample_t));
    return(ALLOC_ERR);
  }
  if((buf2=(sample_t *)malloc(sig->n*sizeof(sample_t))) == NULL) {
    gerror(ALLOC_ERR,"SigDownSample(): cannot allocate %d bytes for buffer #2",
	   sig->n*sizeof(sample_t));
    free(buf1);
    return(ALLOC_ERR);
  }

  for(i=0;i<sig->n;i++)
    buf2[i]=sig->s[i];
  lgth=sig->n;

  while(ndecim < howmany) {
    /*****
     * First, apply the famous FIR filter, with its 110 coefficients ...
     *****/
    for(i=0;i<lgth;i++) { /* for each sample */
      buf1[i]=0.0;
      /****** Compute "from", which is the first filter coefficient to apply *****/
      if(i+(DFILTER_S/2) >= lgth)
	from=i+(DFILTER_S/2)-lgth+1;
      else
	from=0;
      for(j=from;j<DFILTER_S && i+(DFILTER_S/2)-j>=0;j++)
	buf1[i]+=dfilter[j]*buf2[i+(DFILTER_S/2)-j];
    }
    /*****
     * Second, take one sample out of two
     *****/
    for(i=0,j=0;i<lgth;i+=2,j++)
      buf2[j]=buf1[i];
    lgth=j;
    ndecim++;
  }

  free(buf1);
  if((status=CreateSig(out,lgth)) != OK) {
    gerror(status,"SigDownSample(): cannot create output signal");
    free(buf2);
    return(status);
  }
  for(i=0;i<lgth;i++)
    (*out)->s[i]=buf2[i];
  (*out)->n=lgth;
  free(buf2);
  DONE;
}

#undef _spro_global_c_
