/****************************************************/
/*                     spro_io.c                    */
/*                                                  */
/*                   SPro Library                   */
/* ------------------------------------------------ */
/* Guig                                   Apr. 1997 */
/****************************************************/
/*
 * CVS log:
 *
 * $Author: gravier $
 * $Date: 1998/09/17 15:20:33 $
 * $Revision: 1.2 $
 *
 */
/*
 * I/O routines for handling data structures.
 * It includes Create and Free routines, as well as
 * Write and Read.
 * 
 * This module is derived from sigio in SPro v2, but
 * I removed Gnu procedures and added data handling.
 * I also added a procedure to write data to HTK 2.0
 * format, as this version of HTK does not accept
 * ALIEN data files (Isn't that stupid?). Actually,
 * the HTKWrite() routine supports HTK 1.* and 2.*
 * data file formats to avoid storing dcep and ddcep
 * coefficients, as coercion (HCOERCE) doesn't work
 * with alien formats!
 *
 */
#define _spro_io_c_

#include <stdio.h>            /* Standard libraries                   */
#include <stdlib.h>

#include <spro.h>             /* Spro header                          */

#ifdef SPHERE
#include <sp/sphere.h>
#endif

/* -------------------------------------------------- */
/* ----- Private type definition for HTKWrite() ----- */
/* -------------------------------------------------- */
typedef struct {
  long nSamples;
  long sampPeriod;
  short sampSize;
  short parmKind;
} htk_header_t;

#define H_LPC       1
#define H_LPREFC    2
#define H_LPCEPSTRA 3
#define H_MFCC      6
#define H_FBANK     7
#define H_USER      9

#define HASENERGY   0x0040
#define HASNOE      0x0080
#define HASDELTA    0x0100
#define HASZMEAN    0x0800

/* ----------------------------------------------------------- */
/* ----- err_t CreateVector(vector_t **,unsigned long n) ----- */
/* ----------------------------------------------------------- */
err_t CreateVector(vector_t **ptr,unsigned long n)
{
  if(((*ptr)=(vector_t *)malloc(sizeof(vector_t))) == NULL) {
    gerror(ALLOC_ERR,"CreateVector(): cannot allocate %d bytes",sizeof(vector_t));
    return(ALLOC_ERR);
  }
  if(((*ptr)->data=(param_t *)malloc(n*sizeof(param_t))) == NULL)
    {
      free(ptr);
      gerror(ALLOC_ERR,"CreateVector(): cannot allocate %d bytes",n*sizeof(param_t));
      return(ALLOC_ERR);
    }
  (*ptr)->n=n;
  DONE;
}

/* --------------------------------------- */
/* ----- void FreeVector(vector_t *) ----- */
/* --------------------------------------- */
void FreeVector(vector_t *p)
{
free(p->data);
free(p);
}

/* --------------------------------------------------- */
/* ----- err_t CreateSig(sig_t **,unsigned long) ----- */
/* --------------------------------------------------- */
err_t CreateSig(sig_t **s,unsigned long n)
{
  if(((*s)=(sig_t *)malloc(sizeof(sig_t))) == NULL) {
    gerror(ALLOC_ERR,"CreateSig(): cannot allocate %d bytes",sizeof(sig_t));
    return(ALLOC_ERR);
  }
  if(((*s)->s=(sample_t *)malloc(n*sizeof(sample_t))) == NULL) {
    gerror(ALLOC_ERR,"CreateSig(): cannot allocate %d bytes",n*sizeof(sample_t));
    return(ALLOC_ERR);
  }
  (*s)->n=n;
  DONE;
}

/* --------------------------------- */
/* ----- void FreeSig(sig_t *) ----- */
/* --------------------------------- */
void FreeSig(sig_t *s)
{
free(s->s);
free(s);
}

/* --------------------------------------------------------------------------------------- */
/* ----- err_t CreateSPData(data_t **,datakind_t,unsigned long,unsigned long,flag_t) ----- */
/* --------------------------------------------------------------------------------------- */
/*
 * Allocate memory for data structure.
 * nsmp is the number of samples, sdim is the sample dimension. Note that sdim is
 * the dimension of the static coefficients, WITHOUT energy or delta!
 */
err_t CreateSPData(data_t **data,datakind_t kind,unsigned long dim,unsigned long n,flag_t flag)
{
  unsigned long nelem;

  if(((*data)=(data_t *)malloc(sizeof(data_t))) == NULL) {
    gerror(ALLOC_ERR,"CreateSPData(): cannot allocate %d bytes",
	   sizeof(data_t));
    return(ALLOC_ERR);
  }
  (*data)->kind=kind;
  (*data)->dim=dim;
  (*data)->n=n;
  (*data)->flag=flag;
  nelem=DataVecSize(*data)*n;
  if(((*data)->s=(param_t *)malloc(nelem*sizeof(param_t))) == NULL) {
    gerror(ALLOC_ERR,"CreateSPData(): cannot allocate %d bytes",
	   nelem*sizeof(param_t));
    return(ALLOC_ERR);
  }
  DONE;
}

/* ------------------------------------- */
/* ----- void FreeSPData(data_t *) ----- */
/* ------------------------------------- */
void FreeSPData(data_t *p)
{
free(p->s);
free(p);
}

/* ---------------------------------------------- */
/* ----- err_t ReadSPData(data_t **,FILE *) ----- */
/* ---------------------------------------------- */
/*
 * Allocate memory and read data from stream f.
 */
err_t ReadSPData(data_t **data,FILE *f)
{
  datakind_t kind;
  unsigned long dim,n;
  unsigned long nelem;
  flag_t flag;
  err_t status;

 if(fread(&kind,sizeof(datakind_t),1,f) != 1) {
   gerror(LOAD_ERR,"ReadSPData(): cannot read data kind (%d bytes)",
	  sizeof(datakind_t));
   return(LOAD_ERR);
 }
 if(fread(&dim,sizeof(unsigned long),1,f) != 1) {
   gerror(LOAD_ERR,"ReadSPData(): cannot read data dimension (%d bytes)",
	  sizeof(unsigned long));
   return(LOAD_ERR);
 }
 if(fread(&n,sizeof(unsigned long),1,f) != 1) {
   gerror(LOAD_ERR,"ReadSPData(): cannot read data number of elements (%d bytes)",
	  sizeof(unsigned long));
   return(LOAD_ERR);
 }
 if(fread(&flag,sizeof(flag_t),1,f) != 1) {
   gerror(LOAD_ERR,"ReadSPData(): cannot read data qualifiers (%d bytes)",
	  sizeof(flag_t));
   return(LOAD_ERR);
 }
 if((status=CreateSPData(data,kind,dim,n,flag)) != OK) {
   gerror(status,"ReadSPData(): cannot allocate memory for the data structure");
   return(status);
 }
 nelem=DataVecSize(*data)*n;
 if(fread((*data)->s,sizeof(param_t),nelem,f) != nelem) {
   gerror(LOAD_ERR,"ReadSPData(): cannot read data (%d elements = %d bytes)",
	  nelem,nelem*sizeof(param_t));
   FreeSPData(*data);
   return(LOAD_ERR);
 }
 DONE;
}

/* ----------------------------------------------------- */
/* ----- err_t LoadSPData(data_t **,flag_t,FILE *) ----- */
/* ----------------------------------------------------- */
err_t LoadSPData(data_t **data,flag_t cvtflag,FILE *f)
{
  data_t *tmp;
  err_t status;

  if((status=ReadSPData(data,f)) != OK) {
    gerror(status,"LoadSPData(): cannot read data from file");
    return(status);
  }
  if(((*data)->flag | cvtflag) != (*data)->flag) {
    if((status=ConvertData(&tmp,*data,(*data)->flag | cvtflag)) != OK) {
      gerror(status,"LoadSPData(): cannot convert data");
      return(status);
    }
    FreeSPData(*data);
    *data=tmp;
  }
  DONE;
}

/* ---------------------------------------------- */
/* ----- err_t WriteSPData(data_t *,FILE *) ----- */
/* ---------------------------------------------- */
err_t WriteSPData(data_t *data,FILE *f)
{
  int nelem;

  if(fwrite(&(data->kind),sizeof(datakind_t),1,f) != 1) {
    gerror(DUMP_ERR,"WriteSPData(): cannot write data kind (%d bytes)",
	   sizeof(datakind_t));
    return(DUMP_ERR);
  }
  if(fwrite(&(data->dim),sizeof(unsigned long),1,f) != 1) {
    gerror(DUMP_ERR,"WriteSPData(): cannot write data dimension (%d bytes)",
	   sizeof(unsigned long));
    return(DUMP_ERR);
  }
  if(fwrite(&(data->n),sizeof(unsigned long),1,f) != 1) {
    gerror(DUMP_ERR,"WriteSPData(): cannot write data number of elements (%d bytes)",
	   sizeof(unsigned long));
    return(DUMP_ERR);
  }
  if(fwrite(&(data->flag),sizeof(flag_t),1,f) != 1) {
    gerror(DUMP_ERR,"WriteSPData(): cannot write data qualifiers (%d bytes)",
	   sizeof(flag_t));
    return(DUMP_ERR);
  }
  nelem=DataVecSize(data)*data->n;
  if(fwrite(data->s,sizeof(param_t),nelem,f) != nelem) {
   gerror(DUMP_ERR,"WriteSPData(): cannot write data (%d elements = %d bytes)",
	  nelem,nelem*sizeof(param_t));
   return(DUMP_ERR);
  }
  DONE;
}

/* -------------------------------------------------------- */
/* ----- err_t ReadSig(sig_t **,unsigned long,FILE *) ----- */
/* -------------------------------------------------------- */
err_t ReadSig(sig_t **s,unsigned long n,FILE *f)
{
  int i;
  err_t status;
  qsmp_t *quantized;


  if((quantized=(qsmp_t *)malloc(n*sizeof(qsmp_t))) == NULL) {
    gerror(ALLOC_ERR,"ReadSig(): cannot allocate memory for quantized signal (%d bytes)",
	   n*sizeof(qsmp_t));
    return(ALLOC_ERR);
  }
  if((status=CreateSig(s,n)) != OK) {
    gerror(status,"ReadSig(): cannot create %d sample signal",n);
    free(quantized);
    return(status);
  }
  if(fread(quantized,sizeof(qsmp_t),n,f) != n) {
    gerror(LOAD_ERR,"ReadSig(): cannot read signal (%d samples = %d bytes)",n,n*sizeof(qsmp_t));
    FreeSig(*s);
    free(quantized);
    return(LOAD_ERR);
  }
  for(i=0;i<n;i++)
    (*s)->s[i]=(sample_t)quantized[i];
  (*s)->n=n;
  free(quantized);
  DONE;
}

#ifdef SPHERE
/* -------------------------------------------------------- */
/* ----- err_t ReadSphereSig(sig_t **,float *,char *) ----- */
/* -------------------------------------------------------- */
err_t ReadSphereSig(sig_t **s,float *Fe,char *fn)
{
  SP_FILE *sp;
  short **ptr;
  char *coding;
  SP_INTEGER ccount,nsmp,srate;
  long i;
  int flag;
  err_t status;
  
  if((sp=sp_open(fn,"r")) == NULL) {
    gerror(FOPEN_ERR,"ReadSphereSig(): cannot open SPHERE file");
    return(FOPEN_ERR);
  }
  switch((flag=sp_set_data_mode(sp,"SE-PCM:DF-ARRAY"))) {
  case 100: 
    coding=NULL;
    sp_h_get_field(sp,"sample_coding",T_STRING,(void **)(&coding));
    gerror(SPH_CONV_ERR,
	   "ReadSphereSig(): cannot convert from %s to PCM format",
	   coding?"unknown":coding);
    status=SPH_CONV_ERR;
    break;
  case 0:
    status=OK;
    break;
  default:
    gerror(SPH_CONV_ERR,"ReadSPhereSig(): internal sphere library error (sp_set_data_mode()=%d)",flag);
    status=SPH_CONV_ERR;
  }

  if(status != OK) {
    sp_close(sp);
    return(status);
  }

  if(sp_h_get_field(sp,"channel_count",T_INTEGER,(void *)&ccount) > 0) {
    gerror(SPHERE_ERR,"ReadSPhereSig(): cannot read channel_count field");
    sp_close(sp);
    return(SPHERE_ERR);
  }
  if(ccount != 1) {
    gerror(SPH_CHANNEL_WARN,"ReadSphereSig(): mutli-channel sphere file (channel_count= %d)",ccount);
    gerror(SPH_CHANNEL_WARN,"                 loading channel 1");
  }

  if(sp_h_get_field(sp,"sample_count",T_INTEGER,(void *)&nsmp) > 0) {
    gerror(SPHERE_ERR,"ReadSphereSig(): cannot read sample_count field");
    sp_close(sp);
    return(SPHERE_ERR);
  }
  if(sp_h_get_field(sp,"sample_rate",T_INTEGER,(void *)&srate) > 0) {
    gerror(SPHERE_ERR,"ReadSphereSig(): cannot read sample_rate field");
    sp_close(sp);
    return(SPHERE_ERR);
  }
  if((ptr=(short **)sp_data_alloc(sp,nsmp)) == NULL) {
    gerror(ALLOC_ERR,"ReadSPhereSig(): cannot allocate memory for quantized signal");
    sp_close(sp);
    return(ALLOC_ERR);
  }
  if(sp_read_data(ptr,nsmp,sp) != nsmp) {
    gerror(SPHERE_ERR,"ReadSphereSig(): cannot read quantized signal");
    sp_close(sp);
    return(SPHERE_ERR);
  }
  if((status=CreateSig(s,nsmp)) != OK) {
    gerror(status,"ReadSphereSig(): cannot create output signal");
    sp_data_free(sp,ptr);
    sp_close(sp);
    return(status);
  }

  for(i=0;i<nsmp;i++)
    (*s)->s[i]=(sample_t)(ptr[0][i]);
  (*s)->n=nsmp;
  *Fe=(float)srate;
  sp_data_free(sp,ptr);
  sp_close(sp);

  DONE;
}
#endif

/* ------------------------------------------ */
/* ----- err_t WriteSig(sig_t *,FILE *) ----- */
/* ------------------------------------------ */
err_t WriteSig(sig_t *sig,FILE *f)
{
  qsmp_t *quantized;
  int i;

  if((quantized=(qsmp_t *)malloc(sig->n*sizeof(qsmp_t))) == NULL) {
    gerror(ALLOC_ERR,"WriteSig(): cannot allocate memory for quantized signal (%d bytes)",
	   sig->n*sizeof(qsmp_t));
    return(ALLOC_ERR);
  }

  for(i=0;i<sig->n;i++) {
    if(sig->s[i] >= MAX_QSMP) {
      quantized[i]=MAX_QSMP;
      gerror(OVERFLOW_WARN,"WriteSig(): overflow detected");
    } else if(sig->s[i] <= MIN_QSMP) {
      gerror(UNDERFLOW_WARN,"WriteSig(): underflow detected");
      quantized[i]=MIN_QSMP;
    } else
      quantized[i]=(qsmp_t)round(sig->s[i]);
  }
  
  if(fwrite(quantized,sizeof(qsmp_t),sig->n,f) != sig->n) {
    gerror(DUMP_ERR,"WriteSig(): cannot write samples (%d samples = %d bytes)",
	   sig->n,sig->n*sizeof(qsmp_t));
    free(quantized);
    return(DUMP_ERR);
  }
  free(quantized);
  DONE;
}

/* ----------------------------------------------------- */
/* ----- err_t HTKWrite(data_t *,float,int,FILE *) ----- */
/* ----------------------------------------------------- */
err_t HTKWrite(data_t *data,float period,int hversion,FILE *f)
{
  htk_header_t header;
  short htk_kind;
  int nelem;
  float *ptr;
  int i;

  if(sizeof(param_t) != sizeof(float)) {
    gerror(HCOMPAT_ERR,"HTKWrite(): sizeof(param_t) != sizeof(float)");
    return(HCOMPAT_ERR);
  }

  nelem=DataVecSize(data);
  header.nSamples=(long)(data->n);
  header.sampSize=(short)(nelem*sizeof(float));
  header.sampPeriod=(long)(period*10000.0);

  if(hversion == 1)
    switch(data->kind) {
    case FBANK: htk_kind=(short)H_FBANK; break;
    case FBCEPSTRA: htk_kind=(short)H_MFCC; break;
    case LPCEPSTRA: htk_kind=(short)H_LPCEPSTRA; break;
    case LPCOEFF: htk_kind=(short)H_LPC; break;
    case PARCOR: htk_kind=(short)H_LPREFC; break;
    default:
      gerror(NO_HFORMAT_ERR,"HTKWrite(): No equivalent format in HTK 1.4 for %s",
	     kind2str(data->kind));
      return(NO_HFORMAT_ERR);
    } 
  else
    htk_kind=(short)H_USER;

  if(data->flag & WITHE) htk_kind |= HASENERGY;
  if(data->flag & WITHD) htk_kind |= HASDELTA;
  if(data->flag & WITHN) htk_kind |= HASNOE;
  if(hversion == 2 && (data->flag & WITHZ)) htk_kind |= HASZMEAN;
  header.parmKind=htk_kind;

  if(fwrite(&header,sizeof(htk_header_t),1,f) != 1) {
    gerror(DUMP_ERR,"HTKWrite(): cannot write HTK header (%d bytes)",
	   sizeof(htk_header_t));
    return(DUMP_ERR);
  }
  for(i=0;i<data->n;i++) {
    ptr=(float *)DataGetVec(data,i,nelem);
    if(fwrite(ptr,sizeof(float),nelem,f) != nelem) {
      gerror(DUMP_ERR,"HTKWrite(): cannot write %d'th vector (%d bytes)",
	     i+1,nelem*sizeof(float));
      return(DUMP_ERR);
    }
  }
  DONE;
}

#if 0
/*
 * NOT TO BE USED ANYMORE !!!!!!!!!!!!!!!
 */

/* ----------------------------------- */
/* ----- data_t *HTKRead(FILE *) ----- */
/* ----------------------------------- */
data_t *HTKRead(FILE *f)
{
  data_t *data;
  htk_header_t header;
  unsigned long dim;
  unsigned long n;
  flag_t flag=0;
  
  if(fread(&header,sizeof(htk_header_t),1,f) != 1) return(NULL);
  if(header.parmKind & HASDELTA) {
    fprintf(stderr," HTKRead(): cannot read file with delta coefficients\n");
    return(NULL);
  }
  n=header.nSamples;
  dim=header.sampSize/sizeof(float);
  if(header.parmKind & HASENERGY) {flag |= WITHE; dim--;}
  if(header.parmKind & HASZMEAN) flag |= WITHZ;
  if((data=CreateSPData(LPCEPSTRA,dim,n,flag)) == NULL) return(NULL);
}
#endif
