/************************************************************/
/*                        spro_lpc.c                        */
/*                                                          */
/*                       SPro Library                       */
/* -------------------------------------------------------- */
/* Guig                                           Apr. 1997 */
/************************************************************/
/*
 * CVS log:
 *
 * $Author: gravier $
 * $Date: 1999/08/20 12:35:30 $
 * $Revision: 1.2 $
 *
 */
/*
 * Auto regressive modelling procedure for speech processing.
 *
 * Most of the functions were in spro.c (SPro v2.0) before.
 *
 */

#define _spro_lpc_c_

#include <spro.h>

/* --------------------------------------------------------------------------- */
/* ----- err_t ARGModel(sig_t *,float,int,datakind_t,vector_t *,float *) ----- */
/* --------------------------------------------------------------------------- */
/*
 * Perform p-order generalized auto-regressive modelling on sig.
 * Spectral resolution is given by alpha. If alpha=0.0, then a classical
 * procedure is applied.
 * Return a vector of coefficients (LPC, PARCOR or LAR) through v and prediction
 * error through variable sigma.
 * Retrun value is true as long as no error occurs.
 * 
 *
 * Analysis method is Levinson-Durbin direct algorithm which can be found in:
 *       "Traitement de la parole, R. BOITE et M. KUNT", Presses Polytechniques Romandes
 *
 * The general method, using first order dephasing filters, can be found in:
 *       "Analyse spectrale a resolution variable", Ph.D. Thesis,
 *       Christian Chouzenoux, ENST
 *
 */
err_t ARGModel(sig_t *sig,float alpha,int p,datakind_t kind,vector_t *v,float *sigma)
{
  int i,step;
  float r[MAX_LPCORDER+1];             /* auto-correlation vector, from p=0 to p   */
  param_t k[MAX_LPCORDER+1];           /* PARCOR vector                            */
  param_t la[MAX_LPCORDER+1];          /* Local a vector                           */
  param_t tmp[MAX_LPCORDER+1];         /* Buffer to update la[]                    */
  err_t status;

  if(p > MAX_LPCORDER || p <= 0) {
    gerror(LPC_ORDER_ERR,"ARGModel(): invalid analysis order p=%d (max= %d)",
	   p,MAX_LPCORDER);
    return(LPC_ORDER_ERR);
  }
  if(alpha <= -1.0 || alpha >= 1.0) {
    gerror(SPEC_RES_ERR,"ARGModel(): invalid value for spectral deformation (alpha= %.2f)",
	   alpha);
    return(SPEC_RES_ERR);
  }
  if(!(kind == LPCOEFF || kind == PARCOR || kind == LAR)) {
    gerror(PARAM_ID_ERR,"ARGModel(): invalid coefficient kind %s",
	   kind2str(kind));
    return(PARAM_ID_ERR);
  }
  if(v->n < p) {
    gerror(OUT_SIZE_ERR,"ARGModel(): not enough memory allocated to parameter vector");
    return(OUT_SIZE_ERR);
  }

  if(alpha == 0.0)
    AutoCorrel(sig,r,p);
  else if((status=ARGCorrel(sig,alpha,r,p)) != OK) {
    gerror(status,"ARGModel(): cannot compute correlation between filter outputs");
    return(status);
  }

  *sigma=r[0];
  la[0]=1.0;
  k[0]=0.0;

  for(step=1;step<=p;step++) {
    k[step]=0.0;
    if(*sigma != 0.0) {
      for(i=0;i<step;i++)
	k[step]=k[step]-la[i]*r[step-i];
      k[step]/=(*sigma);
    }
    if(fabs(k[step]) > 1)
      gerror(UNSTABLE_FILTER_WARN,"ARGModel(): filter is not stable k[%d]=%.4f",
	     step,k[step]);
    /*****
     * Update LPC vector using tmp vector ...
     *****/
    for(i=1;i<step;i++)
      tmp[i]=la[i]+k[step]*la[step-i];
    for(i=1;i<step;i++)
      la[i]=tmp[i];
    la[step]=k[step];
    *sigma=(*sigma)*(1.0-k[step]*k[step]);
  }

  /*****
   * Now fill in coeff according to what was required
   *****/
  switch(kind) {
  case LPCOEFF:
    for(i=0;i<p;i++)
      v->data[i]=la[i+1];
    break;
  case PARCOR:
    for(i=0;i<p;i++)
      v->data[i]=k[i+1];
    break;
  case LAR:
    for(i=0;i<p;i++)
      v->data[i]=10.0*log10((1.0+k[i+1])/(1.0-k[i+1]));
    break;
  default:; /* To avoid some gcc warnings */
  }
  *sigma=(float)sqrt(*sigma);
  DONE;
}

/* ----------------------------------------------------------------------------------- */
/* ----- err_t SigARInverse(sig_t *,vector_t *,datakind_t,float,int,int,sig_t *) ----- */
/* ----------------------------------------------------------------------------------- */
/*
 * Apply (inverse) filtering A(z)/sigma to signal x with coefficients in vector v
 * between samples st and en. Residual signal y must be allocated before.
 * If (st-i >= 0), then x[st-i] and y[st-i] are used as filter memory.
 *
 * NOTE: No control is performed on memory allocated to returned signal y.
 */
err_t SigARInverse(sig_t *y,vector_t *v,datakind_t kind,float sigma,int st,int en,sig_t *x)
{
  int i,j;                    /* the usual loopers            */
  vector_t *a;                /* LPC vector                   */
  err_t status;

  if(v->n > MAX_LPCORDER || v->n <= 0) {
    gerror(LPC_ORDER_ERR,"SigARInverse(): invalid prediction order p=%d (max= %d)",
	   v->n,MAX_LPCORDER);
    return(LPC_ORDER_ERR);
  }
  if(st < 0 || st > en || en >= x->n) {
    gerror(RANGE_ERR,"SigARInverse(): invalid start and/or end sample(s) (st= %d, en= %d)",
	   st,en);
    return(RANGE_ERR);
  }
  if(kind != LPCOEFF) {
    if((status=CreateVector(&a,v->n)) != OK) {
      gerror(status,"SigARInverse(): cannot create LPCOEFF vector");
      return(status);
    }
    if((status=ARConvert(v,kind,LPCOEFF,a)) != OK) {
      gerror(status,"SigARInverse(): cannot convert %s coeffs to LPCOEFF",
	     kind2str(kind));
      FreeVector(a);
      return(status);
    }
  }
  else
    a=v;
  /*****
   * Now, run inverse filtering form sample st to sample en.
   *****/
  if(sigma == 0.0) {
    gerror(NULL_FRAME_WARN,"SigARInverse(): null filter gain sigma (returning zeros)");
    for(i=st;i<=en;i++)
      y->s[i]=(sample_t)0.0;
  }
  else {
    for(i=st;i<=en;i++) {
      y->s[i]=x->s[i];
      for(j=1;j <= a->n && i-j > 0;j++)
	y->s[i] += x->s[i-j]*a->data[j-1];
      y->s[i] /= sigma;
    }
  }
  y->n=en+1;
  if(kind != LPCOEFF)
    FreeVector(a);
  DONE;
}

/* ---------------------------------------------------------------------------------- */
/* ----- err_t SigARFilter(sig_t *,vector_t *,datakind_t,float,int,int,sig_t *) ----- */
/* ---------------------------------------------------------------------------------- */
/*
 * Implementation of an AR(p) filter.
 * The filter looks somewhat like this:
 *                          ---------------
 *                          |    sigma    |
 *            x->s[i] ----->|    -----    |-----> y->s[i]    i=st,...,en
 *                          |     A(z)    |
 *                          ---------------
 * Nice scheme, isn't it?
 *
 * If (st-i >= 0), then x[st-i] and y[st-i] are used as filter memory.
 *
 * NOTE: No control is performed on memory allocated to returned signal y.
 */
err_t SigARFilter(sig_t *y,vector_t *v,datakind_t kind,float sigma,int st,int en,sig_t *x)
{
  int i,j;                    /* the user loopers             */
  vector_t *a;                /* LPC vector                   */
  err_t status;

  if(v->n > MAX_LPCORDER || v->n <= 0) {
    gerror(LPC_ORDER_ERR,"SigARFilter(): invalid prediction order p=%d (max= %d)",
	   v->n,MAX_LPCORDER);
    return(LPC_ORDER_ERR);
  }
  if(st < 0 || st > en || en >= x->n) {
    gerror(RANGE_ERR,"SigARInverse(): invalid start and/or end sample(s) (st= %d, en= %d)",
	   st,en);
    return(RANGE_ERR);
  }

  if(kind != LPCOEFF) {
    if((status=CreateVector(&a,v->n)) != OK) {
      gerror(status,"SigARFilter(): cannot create LPCOEFF vector");
      return(status);
    }
    if((status=ARConvert(v,kind,LPCOEFF,a)) != OK) {
      gerror(status,"SigARFilter(): cannot convert %s coeffs to LPCOEFF",
	     kind2str(kind));
      FreeVector(a);
      return(status);
    }
  }
  else
    a=v;
  /*****
   * Now, run direct filtering ...
   *****/
  for(i=st;i<=en;i++) {
    y->s[i]=sigma*x->s[i];
    for(j=1;j<=a->n && i-j > 0;j++)
      y->s[i] -= y->s[i-j]*a->data[j-1];
  }
  y->n=en+1;

  if(kind != LPCOEFF)
    FreeVector(a);
  DONE;
}


/* ------------------------------------------------------------------------ */
/* ----- err_t ARConvert(vector_t *,datakind_t,datakind_t,vector_t *) ----- */
/* ------------------------------------------------------------------------ */
/*
 * Perform conversion: LAR <-----> PARCOR <-----> LPC
 * v1 (resp. v2) and t1 (resp. t2) are the source (resp. target) vector and
 * coefficient type.
 * Return value is false if conversion is impossible.
 * Else, the function returns true and v2 contains the new coefficients.
 *
 */
err_t ARConvert(vector_t *v1,datakind_t t1,datakind_t t2,vector_t *v2)
{
  param_t buf[MAX_LPCORDER];  /* buffer                                   */

  if(v1->n > MAX_LPCORDER || v1->n <= 0) {
    gerror(LPC_ORDER_ERR,"ARConvert(): invalid prediction order p=%d (max= %d)",
	   v1->n,MAX_LPCORDER);
    return(LPC_ORDER_ERR);
  }
  if(v2->n < v1->n) {
    gerror(OUT_SIZE_ERR,"ARConvert(): not enough memory allocated to output vector");
    return(OUT_SIZE_ERR);
  }
  if(t1 !=  LPCOEFF  && t1 != PARCOR && t1 != LAR) {
    gerror(PARAM_ID_ERR,"ARConvert(): source kind %s is not valid",kind2str(t1));
    return(PARAM_ID_ERR);
  }
  if(t2 !=  LPCOEFF  && t2 != PARCOR && t2 != LAR) {
    gerror(PARAM_ID_ERR,"ARConvert(): target kind %s is not valid",kind2str(t1));
    return(PARAM_ID_ERR);
  }
  if(t1 == t2) {
    gerror(CONVERT_WARN,"ARConvert(): target and source kind are the same");
    return(CONVERT_WARN);
  }

  v2->n=v1->n;
  if(t1 == LAR) {
    if(t2 == PARCOR)
      g2k(v1->data,v1->n,v2->data);
    else { /* t2 == LPCOEFF */
      g2k(v1->data,v1->n,buf);
      k2a(buf,v1->n,v2->data);
    }
  }
  else if(t1 == PARCOR) {
    if(t2 == LPCOEFF)
      k2a(v1->data,v1->n,v2->data);
    else /* t2 == LAR */
      k2g(v1->data,v1->n,v2->data);
  }
  else { /* t1 == LPCOEFF */
    if(t2 == PARCOR)
      a2k(v1->data,v1->n,v2->data);
    else { /* t2 == LAR */
      a2k(v1->data,v1->n,buf);
      k2g(buf,v1->n,v2->data);
    }
  }
  DONE;
}

/* ------------------------------------------------------- */
/* ----- err_t LPCepstrum(vector_t *,int,vector_t *) ----- */
/* ------------------------------------------------------- */
/*
 * Return p cepstral coefficients into vector c, computed
 * from prediction vector a.
 *
 * The formula is (assuming {a[1],...,a[N]}):
 *    c[i]=-a[i]+(1/i)*sum for j=1 to i-1 of((i-j)*a[j]*c[i-j]) i=1,..,p
 *
 * This is what is implemented with arrays indexed from 0 to N-1.
 *
 */
err_t LPCepstrum(vector_t *a,int p,vector_t *c)
{
  int j,i;

  if(a->n < p || p < 2) {
    gerror(
	   LPCC_ORDER_ERR,
	   "LPCepstrum(): cepstral analysis order (%d) incompatible with LP order (%d)",
	   p,a->n);
    return(LPCC_ORDER_ERR);
  }
  if(c->n < p) {
    gerror(OUT_SIZE_ERR,"LPCepstrum(): not enough memory in output vector");
    return(OUT_SIZE_ERR);
  }

  for(i=1;i<=p;i++) {
    c->data[i-1]=(param_t)0.0;
    for(j=1;j<i;j++)
      c->data[i-1] -= ((float)(i-j)*a->data[j-1]*c->data[i-j-1]);
    c->data[i-1] *= (1.0/(float)i);
    c->data[i-1] -= (a->data[i-1]);
  }
  c->n=p;
  DONE;
}
    
        /******************************/
        /***** PRIVATE PROCEDURES *****/
        /******************************/

/* ------------------------------------------------------- */
/* ----- void g2k(param_t *,unsigned long,param_t *) ----- */
/* ------------------------------------------------------- */
void g2k(param_t *g,unsigned long p,param_t *k)
{
  int i;
  float alpha;
 
  for(i=0;i<p;i++) {
    alpha=pow(10,g[i]/10.0);
    k[i]=(param_t)(alpha-1.0)/(alpha+1.0);
  }
}

/* ------------------------------------------------------- */
/* ----- void k2g(param_t *,unsigned long,param_t *) ----- */
/* ------------------------------------------------------- */
void k2g(param_t *k,unsigned long p,param_t *g)
{
  int i;

  for(i=0;i<p;i++)
    g[i]=10.0*log10((1.0+k[i])/(1.0-k[i]));
}

/* ------------------------------------------------------- */
/* ----- void k2a(param_t *,unsigned long,param_t *) ----- */
/* ------------------------------------------------------- */
/*
 * The trick is to compute new a (prediction order i) from
 * old a (prediction order i-1): The formulae are:
 *  order 1:       
 *     a1[1]=k1
 *  from i-1 to i: 
 *     ai[j]=ai-1[j]+ki*ai-1[i-j] j=1,...,i-1
 *     ai[i]=ki
 * The operation is repeated from order 1 to p.
 * Arrays are indexed from 0 to 1 but this procedure 
 * implements those formulae.
 */
void k2a(param_t *k,unsigned long p,param_t *a)
{
  int i,j;
  param_t olda[MAX_LPCORDER];

  for(i=1;i<=p;i++) {
    for(j=0;j<i-1;j++)
      a[j]=olda[j]+k[i-1]*olda[i-j-2];
    a[i-1]=k[i-1];
    for(j=0;j<i;j++)
      olda[j]=a[j];
  }
}

/* ------------------------------------------------------- */
/* ----- void a2k(param_t *,unsigned long,param_t *) ----- */
/* ------------------------------------------------------- */
/*
 * The trick is now to compute previous a (order i-1) from
 * current one (order i) as ki-1=ai-1[i-1]. It gives something
 * like these formulae:
 * From order i to i-1:
 *     ki=ai[i]
 *     ai-1[j]=(ai[j]-ki*ai[i-j])/(1-ki*ki) j=1,...,i-1
 * The operation is repeated from order p to 1.
 * Arrays are indexed from 0 to 1 but this procedure 
 * implements those formulae.
 */

void a2k(param_t *a,unsigned long p,param_t *k)
{
  int i,j;
  float x;
  param_t curra[MAX_LPCORDER];
  param_t preva[MAX_LPCORDER];
  
  for(i=0;i<p;i++) curra[i]=a[i];
  for(i=p;i>=1;i--) {
    k[i-1]=curra[i-1];
    x=1-k[i-1]*k[i-1];
    for(j=0;j<i-1;j++)
      preva[j]=(curra[j]-k[i-1]*preva[i-j-2])/x;
    for(j=0;j<i-1;j++)
      curra[i]=preva[i];
  }
}

/* ------------------------------------------------ */
/* ----- void AutoCorrel(sig_t *,float *,int) ----- */
/* ------------------------------------------------ */
/*****
 * Calculate r[p]=E[x[n]x[n-p]], for p=0,...,order
 *
 * NOTE: isig is assumed to be centered ...
 *
*****/
void AutoCorrel(sig_t *isig,float *r,int order)
{
  int i,p;

  for(p=0;p<=order;p++) {
    r[p]=0.0;
    for(i=p;i<isig->n;i++)
      r[p]+=isig->s[i]*isig->s[i-p];
  }
}

/* ------------------------------------------------------ */
/* ----- err_t ARGCorrel(sig_t *,float,float *,int) ----- */
/* ------------------------------------------------------ */
/*****
 * Generalized "correlation sequence".
 * First apply the corrective filter mu(z)=(1-alpha^2)/(1-(alpha.z-1))^2.
 * Filter the signal by H(z) p times and estimate correlation ...
*****/
err_t ARGCorrel(sig_t *sig,float alpha,float *r,int p)
{
  int i,j;
  bool_t noerror=true;
  sig_t outputs[MAX_LPCORDER+1];

  /*****
   * Apply corrective filter ...
   *****/
  if((outputs[0].s=(sample_t *)malloc(sig->n*sizeof(sample_t))) == NULL) {
    gerror(ALLOC_ERR,"ARGCorrel(): cannot allocate memory for first buffer (%d bytes)",
	   sig->n*sizeof(sample_t));
    return(ALLOC_ERR);
  }
  outputs[0].n=sig->n;
  outputs[0].s[0]=sig->s[0]*(1.0-alpha*alpha);
  outputs[0].s[1]=(1.0-alpha*alpha)*sig->s[1]+2.0*alpha*outputs[0].s[0];
  for(i=2;i<sig->n;i++)
    outputs[0].s[i]=(1-alpha*alpha)*sig->s[i]+2.0*alpha*outputs[0].s[i-1]-alpha*alpha*outputs[0].s[i-2];

  /*****
   * Succesively apply the H(z) first order filter ...
   * x ---> H(z) ---> y: y[n]=-alpha.x[n]+x[n-1]+alpha.y[n-1]
   * Here, y is outputs[i] and x is the output from previous filter, i.e.
   * outputs[i-1] ...
   * Let's go hoping it works out ...
   *****/
  for(i=1;i<=p && noerror;i++)
    if((outputs[i].s=(sample_t *)malloc(sig->n*sizeof(sample_t))) == NULL) {
      gerror(ALLOC_ERR,"ARGCorrel(): cannot allocate memory for %d'th buffer (%d bytes)",
	     i,sig->n*sizeof(sample_t));
      for(j=i-1;j >= 0;j--)
	free(outputs[i].s);
      return(ALLOC_ERR);
    }
    else {
      outputs[i].n=sig->n;
      outputs[i].s[0]=-alpha*outputs[i-1].s[0];
      for(j=1;j<outputs[i].n;j++)
	outputs[i].s[j]=-alpha*outputs[i-1].s[j]+outputs[i-1].s[j-1]+alpha*outputs[i].s[j-1];
    }
  /*****
   * We now have to compute the correlation vector "r".
   *****/
  for(i=0;i<=p;i++) {
    r[i]=0.0;
    for(j=0;j<sig->n;j++)
      r[i]=r[i]+outputs[0].s[j]*outputs[i].s[j];
  }
  /****
   * Finally, free what's been allocated and leave ...
   *****/
  for(i=0;i<=p;i++)
    free(outputs[i].s);
  DONE;
}

#undef _spro_lpc_c_
