/************************************************************/
/*                        sprognu.c                         */
/*                                                          */
/*                      SPro Sources                        */
/* -------------------------------------------------------- */
/* Guig                                           Apr. 1997 */
/************************************************************/
/*
 * CVS log:
 *
 * $Author: gravier $
 * $Date: 1998/07/21 14:29:23 $
 * $Revision: 1.1.1.1 $
 *
 */
/*
 * Provides several procedures for exporting files in gnuplot
 * format.
 */

#include <stdio.h>
#include <stdlib.h>

#include <gutil.h>
#include <gerror.h>
#include <errcode.h>
#include <spro.h>

#define FFT_NPTS  512
#define DB_BASE   (float)10000.0

/* ------------------------------------------------- */
/* ----- void GnuARSpectrum(vector_t *,char *) ----- */
/* ------------------------------------------------- */
/*
 * Compute the transfer function of the filter 1/A(z)
 * defined by vector a and writes it to file fname.
 * If alpha is not 0.0, the transfer function is given on the
 * transformed frequency axis omega', else it is given in
 * the original frequency axis omega.
 */
void GnuARSpectrum(vector_t *a,float freq,float alpha,char *fname)
{
  int i;
  FILE *f;
  sig_t *buf;
  float re,im,om;
  vector_t *module,*phase;
  extern float ThetaInv(float,float);
  err_t status;

  if((f=fopen(fname,"w")) != NULL) {
    if((status=CreateVector(&module,FFT_NPTS/2)) != OK) {
      gerror(status,"GnuARSpectrum(): cannot allocate memory for module vector");
      fclose(f);
      return;
    }
    if((status=CreateVector(&phase,FFT_NPTS/2)) != OK) {
      gerror(status,"GnuARSpectrum(): cannot allocate memory for phase vector");
      FreeVector(module);
      fclose(f);
      return;
    }
    if((status=CreateSig(&buf,a->n+1)) != OK) {
      gerror(status,"GnuARSpectrum(): cannot allocate memory");
      FreeVector(module); FreeVector(phase);
      fclose(f);
      return;
    }
    buf->s[0]=1.0;
    for(i=1;i<=a->n;i++)
      buf->s[i]=(sample_t)(a->data[i-1]);
    if((status=FFTransform(buf,FFT_NPTS,module,phase)) != OK) {
      gerror(status,"GnuARSpectrum(): cannot run FFT analysis");
      FreeVector(module); FreeVector(phase);
      fclose(f);
      return;
    }
    for(i=0;i<module->n;i++) {
      re=cos(phase->data[i])/module->data[i];
      im=-sin(phase->data[i])/module->data[i];
      om=i*M_PI/(float)(module->n-1);
      if(alpha == 0)
	fprintf(f,"%.4f %f\n",(float)i*freq/(2.0*(module->n-1)),(float)(10.0*log10((re*re+im*im)/DB_BASE)));
      else
	fprintf(f,"%.4f %f\n",ThetaInv(om,alpha)*freq/(2.0*M_PI),(float)(10.0*log10((re*re+im*im)/DB_BASE)));
    }
    FreeVector(module);
    FreeVector(phase);
    fclose(f);
  }
}

/* -------------------------------------------------- */
/* ----- void GnuSpectrum(sig_t *,float,char *) ----- */
/* -------------------------------------------------- */
void GnuSpectrum(sig_t *sig,float freq,char *fname)
{
  FILE *f;
  int i;
  int npts;
  vector_t *module;
  err_t status;

  if((f=fopen(fname,"w")) != NULL) {
    npts=sig->n < FFT_NPTS?FFT_NPTS:sig->n;
    if((status=CreateVector(&module,npts/2)) != OK) {
      gerror(status,"GnuSpectrum(): cannot create vector");
      return;
    }
    if((status=FFTransform(sig,npts,module,NULL)) != OK) {
      gerror(status,"GnuSpectrum(): cannot run FFT");
      FreeVector(module);
      return;
    }
    for(i=0;i<npts/2;i++)
      fprintf(f,"%f %f\n",(float)i*freq/(2.0*(module->n-1)),(float)(10.0*log10((module->data[i]*module->data[i])/DB_BASE)));
    FreeVector(module);
    fclose(f);
  }
}

/* ----------------------------------------- */
/* ----- void GnuData(data_t *,FILE *) ----- */
/* ----------------------------------------- */
void GnuData(data_t *data,FILE *f)
{
  int i,j;
  unsigned long vecsize;

  vecsize=DataVecSize(data);
  for(i=0;i<data->n;i++) {
    for(j=0;j<vecsize;j++)
      fprintf(f,"%f ",data->s[i*vecsize+j]);
    fprintf(f,"\n");
  }
}

/* --------------------------------------------- */
/* ----- void GnuSig(sig_t *,float,char *) ----- */
/* --------------------------------------------- */
void GnuSig(sig_t *sig,float srate,char *fname)
{
  FILE *f;
  int i;

  if((f=fopen(fname,"w")) != NULL) {
    for(i=0;i<sig->n;i++)
      fprintf(f,"%f %f\n",i*1000.0/(float)srate,sig->s[i]);
    fclose(f);
  }
}
